"""
This file contains the 'double_buffered_scratchpad' class that handles memory module simulations of
double buffered SRAMs.
"""

import time
import os
import numpy as np
from tqdm import tqdm

from scalesim.scale_config import scale_config as cfg
from scalesim.topology_utils import topologies as topo
from scalesim.memory.read_buffer import read_buffer as rdbuf
from scalesim.memory.read_buffer_estimate_bw import ReadBufferEstimateBw as rdbuf_est
from scalesim.memory.read_port import read_port as rdport
from scalesim.memory.write_buffer import write_buffer as wrbuf
from scalesim.memory.write_port import write_port as wrport

class double_buffered_scratchpad:
    """
    Class which runs the memory simulation of double buffered scratchpad memories (SRAMs). The
    double buffering helps to hide the DRAM latency when the SRAM is servicing requests from the
    systolic array using one of the buffers while the other buffer prefetches from the DRAM.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        self.layer_id = 0
        self.ifmap_buf = rdbuf()
        self.filter_buf = rdbuf()
        self.ofmap_buf =wrbuf()

        self.ifmap_port = rdport()
        self.filter_port = rdport()
        self.ofmap_port = wrport()
        self.config = cfg()
        self.topo = topo()

        self.verbose = True

        self.ifmap_trace_matrix = np.zeros((1,1), dtype=int)
        self.filter_trace_matrix = np.zeros((1,1), dtype=int)
        self.ofmap_trace_matrix = np.zeros((1,1), dtype=int)

        # Metrics to gather for generating run reports
        self.total_cycles = 0
        self.compute_cycles = 0
        self.stall_cycles = 0

        self.avg_ifmap_dram_bw = 0
        self.avg_filter_dram_bw = 0
        self.avg_ofmap_dram_bw = 0

        self.ifmap_sram_start_cycle = 0
        self.ifmap_sram_stop_cycle = 0
        self.filter_sram_start_cycle = 0
        self.filter_sram_stop_cycle = 0
        self.ofmap_sram_start_cycle = 0
        self.ofmap_sram_stop_cycle = 0

        self.ifmap_dram_start_cycle = 0
        self.ifmap_dram_stop_cycle = 0
        self.ifmap_dram_reads = 0
        self.filter_dram_start_cycle = 0
        self.filter_dram_stop_cycle = 0
        self.filter_dram_reads = 0
        self.ofmap_dram_start_cycle = 0
        self.ofmap_dram_stop_cycle = 0
        self.ofmap_dram_writes = 0

        self.estimate_bandwidth_mode = False
        self.traces_valid = False
        self.params_valid_flag = True
        self.use_ramulator_trace = self.config.get_ramulator_trace()

        self.using_ifmap_custom_layout = False
        self.using_filter_custom_layout = False

    #
    def set_params(self,
                   layer_id=0,
                   verbose=True,
                   estimate_bandwidth_mode=False,
                   word_size=1,
                   ifmap_buf_size_bytes=2, filter_buf_size_bytes=2, ofmap_buf_size_bytes=2,
                   rd_buf_active_frac=0.5, wr_buf_active_frac=0.5,
                   ifmap_backing_buf_bw=1, filter_backing_buf_bw=1, ofmap_backing_buf_bw=1,
                   ifmap_sram_bank_num=1, ifmap_sram_bank_port=2, filter_sram_bank_num=1, filter_sram_bank_port=2,
                   using_ifmap_custom_layout=False, using_filter_custom_layout=False,
                   config=cfg(), topo=topo()
                   ):

        """
        Method to set the double buffered memory simulation parameters for housekeeping.
        """
        self.layer_id = layer_id
        self.topo = topo
        self.config = config
        self.use_ramulator_trace = config.get_ramulator_trace()

        self.estimate_bandwidth_mode = estimate_bandwidth_mode

        if self.estimate_bandwidth_mode:
            self.ifmap_buf = rdbuf_est()
            self.filter_buf = rdbuf_est()

            self.ifmap_buf.set_params(backing_buf_obj=self.ifmap_port,
                                      total_size_bytes=ifmap_buf_size_bytes,
                                      word_size=word_size,
                                      active_buf_frac=rd_buf_active_frac,
                                      backing_buf_default_bw=ifmap_backing_buf_bw,
                                      use_ramulator_trace=self.use_ramulator_trace
                                      )

            self.filter_buf.set_params(backing_buf_obj=self.filter_port,
                                       total_size_bytes=filter_buf_size_bytes,
                                       word_size=word_size,
                                       active_buf_frac=rd_buf_active_frac,
                                       backing_buf_default_bw=filter_backing_buf_bw,
                                       use_ramulator_trace=self.use_ramulator_trace
                                       )
        else:
            self.ifmap_buf = rdbuf()
            self.filter_buf = rdbuf()
            
            if self.use_ramulator_trace == True:
                root_path = os.getcwd()
                #topology_file = self.topo.split('.')[0]
                topology_file =''
                ifmap_dram_trace = (root_path+"/results/"+topology_file+"_ifmapFile"+str(layer_id)+".npy")
                filter_dram_trace = (root_path+"/results/"+topology_file+"_filterFile"+str(layer_id)+".npy")
                ofmap_dram_trace = (root_path+"/results/"+topology_file+"_ofmapFile"+str(layer_id)+".npy")
                self.ifmap_port.def_params(config = self.config, latency_file=ifmap_dram_trace)
                self.filter_port.def_params(config = self.config, latency_file=filter_dram_trace)
                self.ofmap_port.def_params(config=self.config, latency_file=ofmap_dram_trace)

            self.ifmap_buf.set_params(backing_buf_obj=self.ifmap_port,
                                      total_size_bytes=ifmap_buf_size_bytes,
                                      word_size=word_size,
                                      active_buf_frac=rd_buf_active_frac,
                                      backing_buf_bw=ifmap_backing_buf_bw,
                                      num_bank=ifmap_sram_bank_num,
                                      num_port=ifmap_sram_bank_port,
                                      enable_layout_evaluation=using_ifmap_custom_layout,
                                      use_ramulator_trace=self.use_ramulator_trace
                                      )

            self.filter_buf.set_params(backing_buf_obj=self.filter_port,
                                       total_size_bytes=filter_buf_size_bytes,
                                       word_size=word_size,
                                       active_buf_frac=rd_buf_active_frac,
                                       backing_buf_bw=filter_backing_buf_bw,
                                       num_bank=filter_sram_bank_num,
                                       num_port=filter_sram_bank_port,
                                       enable_layout_evaluation=using_filter_custom_layout,
                                       use_ramulator_trace=self.use_ramulator_trace
                                       )

        self.ofmap_buf.set_params(backing_buf_obj=self.ofmap_port,
                                  total_size_bytes=ofmap_buf_size_bytes,
                                  word_size=word_size,
                                  active_buf_frac=wr_buf_active_frac,
                                  backing_buf_bw=ofmap_backing_buf_bw)

        self.verbose = verbose

        self.using_ifmap_custom_layout = using_ifmap_custom_layout  
        self.using_filter_custom_layout = using_filter_custom_layout  
        self.params_valid_flag = True


    #
    def set_read_buf_prefetch_matrices(self,
                                       ifmap_prefetch_mat=np.zeros((1,1)),
                                       filter_prefetch_mat=np.zeros((1,1))
                                       ):
        """
        Method to read ifmap and filter prefetch matrices generated in the compute simulation.
        """

        self.ifmap_buf.set_fetch_matrix(ifmap_prefetch_mat)
        self.filter_buf.set_fetch_matrix(filter_prefetch_mat)

    #
    def reset_buffer_states(self):
        """
        Method to reset ifmap, filter and ofmap SRAMs.
        """

        self.ifmap_buf.reset()
        self.filter_buf.reset()
        self.ofmap_buf.reset()

    # The following are just shell methods for users to control each mem individually
    def service_ifmap_reads(self,
                            incoming_requests_arr_np,   # 2D array with the requests
                            incoming_cycles_arr):
        """
        Method to service ifmap read requests coming from systolic array.
        """
        out_cycles_arr_np = self.ifmap_buf.service_reads(incoming_requests_arr_np,
                                                         incoming_cycles_arr)

        return out_cycles_arr_np

    #
    def service_filter_reads(self,
                            incoming_requests_arr_np,   # 2D array with the requests
                            incoming_cycles_arr):
        """
        Method to service filter read requests coming from systolic array.
        """
        out_cycles_arr_np = self.filter_buf.service_reads(incoming_requests_arr_np,
                                                          incoming_cycles_arr)

        return out_cycles_arr_np

    #
    def service_ofmap_writes(self,
                             incoming_requests_arr_np,  # 2D array with the requests
                             incoming_cycles_arr):
        """
        Method to service ofmap write requests coming from systolic array.
        """

        out_cycles_arr_np = self.ofmap_buf.service_writes(incoming_requests_arr_np,
                                                          incoming_cycles_arr, 1)

        return out_cycles_arr_np

    #
    def service_memory_requests(self, ifmap_demand_mat, filter_demand_mat, ofmap_demand_mat):
        """
        Method to run the memory simulation of ifmap, filter and ofmap SRAMs together and generate
        the traces.
        """
        assert self.params_valid_flag, 'Memories not initialized yet'

        ofmap_lines = ofmap_demand_mat.shape[0]

        self.total_cycles = 0
        self.stall_cycles = 0

        ifmap_hit_latency = self.ifmap_buf.get_hit_latency()
        filter_hit_latency = self.filter_buf.get_hit_latency()

        ifmap_serviced_cycles = []
        filter_serviced_cycles = []
        ofmap_serviced_cycles = []

        pbar_disable = not self.verbose
        for i in tqdm(range(ofmap_lines), disable=pbar_disable):

            cycle_arr = np.zeros((1,1)) + i + self.stall_cycles

            ifmap_demand_line = ifmap_demand_mat[i, :].reshape((1,ifmap_demand_mat.shape[1]))
            ifmap_cycle_out = \
                self.ifmap_buf.service_reads(incoming_requests_arr_np=ifmap_demand_line,
                                             incoming_cycles_arr=cycle_arr)
            ifmap_serviced_cycles += [ifmap_cycle_out[0]]
            ifmap_stalls = ifmap_cycle_out[0] - cycle_arr[0] - ifmap_hit_latency

            filter_demand_line = filter_demand_mat[i, :].reshape((1, filter_demand_mat.shape[1]))
            filter_cycle_out = \
                self.filter_buf.service_reads(incoming_requests_arr_np=filter_demand_line,
                                              incoming_cycles_arr=cycle_arr)
            filter_serviced_cycles += [filter_cycle_out[0]]
            filter_stalls = filter_cycle_out[0] - cycle_arr[0] - filter_hit_latency

            ofmap_demand_line = ofmap_demand_mat[i, :].reshape((1, ofmap_demand_mat.shape[1]))
            ofmap_cycle_out = \
                self.ofmap_buf.service_writes(incoming_requests_arr_np=ofmap_demand_line,
                                              incoming_cycles_arr_np=cycle_arr)
            ofmap_serviced_cycles += [ofmap_cycle_out[0]]
            ofmap_stalls = ofmap_cycle_out[0] - cycle_arr[0]

            self.stall_cycles += int(max(ifmap_stalls[0], filter_stalls[0], ofmap_stalls[0]))
            #self.stall_cycles += ifmap_stalls[0] + filter_stalls[0] + ofmap_stalls[0]

        if self.estimate_bandwidth_mode:
            # IDE shows warning as complete_all_prefetches is not implemented in read_buffer class
            # It's harmless since read_buffer_estimate_bw is instantiated in estimate bandwidth mode
            self.ifmap_buf.complete_all_prefetches()
            self.filter_buf.complete_all_prefetches()

        self.ofmap_buf.empty_all_buffers(ofmap_serviced_cycles[-1])

        # Prepare the traces
        ifmap_services_cycles_np = \
            np.asarray(ifmap_serviced_cycles).reshape((len(ifmap_serviced_cycles), 1))
        self.ifmap_trace_matrix = np.concatenate((ifmap_services_cycles_np, ifmap_demand_mat),
                                                 axis=1)

        filter_services_cycles_np = \
            np.asarray(filter_serviced_cycles).reshape((len(filter_serviced_cycles), 1))
        self.filter_trace_matrix = np.concatenate((filter_services_cycles_np, filter_demand_mat),
                                                  axis=1)

        ofmap_services_cycles_np = \
            np.asarray(ofmap_serviced_cycles).reshape((len(ofmap_serviced_cycles), 1))
        self.ofmap_trace_matrix = np.concatenate((ofmap_services_cycles_np, ofmap_demand_mat),
                                                 axis=1)
        #self.total_cycles = int(ofmap_serviced_cycles[-1][0])
        ## Probable fault in sanity check
        self.total_cycles = int(max(ofmap_serviced_cycles))

        # END of serving demands from memory
        self.traces_valid = True

    # This is the trace computation logic of this memory system
    # Anand: This is too complex, perform the serve cycle by cycle for the requests
    def service_memory_requests_old(self, ifmap_demand_mat, filter_demand_mat, ofmap_demand_mat):
        """
        This is the trace computation logic of this memory system.
        """
        # TODO: assert sanity check
        assert self.params_valid_flag, 'Memories not initialized yet'

        # Logic:
        # Stalls can occur in both read and write portions and interfere with each other
        # We mitigate interference by picking a window in which there are no write stall,
        # ie, there is sufficient free space in the write buffer

        # The three demand mats have the same shape though
        ofmap_lines_remaining = ofmap_demand_mat.shape[0]
        start_line_idx = 0
        end_line_idx = 0

        first = True
        cycle_offset = 0
        self.total_cycles = 0
        self.stall_cycles = 0

        # Status bar
        pbar_disable = not self.verbose #or True
        pbar = tqdm(total=ofmap_lines_remaining, disable=pbar_disable)

        avg_read_time_series = []

        while ofmap_lines_remaining > 0:
            loop_start_time = time.time()
            ofmap_free_space = self.ofmap_buf.get_free_space()

            # Find the number of lines till the ofmap_free_space is filled up
            count = 0
            while not count > ofmap_free_space:
                this_line = ofmap_demand_mat[end_line_idx]
                for elem in this_line:
                    if not elem == -1:
                        count += 1

                if not count > ofmap_free_space:
                    end_line_idx += 1
                    # Limit check
                    if not end_line_idx < ofmap_demand_mat.shape[0]:
                        end_line_idx = ofmap_demand_mat.shape[0] - 1
                        count = ofmap_free_space + 1
                else:   # Send request with minimal data ie one line of the requests
                    end_line_idx += 1
            # END of line counting

            num_lines = end_line_idx - start_line_idx + 1
            this_req_cycles_arr = [int(x + cycle_offset) for x in range(num_lines)]
            this_req_cycles_arr_np = np.asarray(this_req_cycles_arr).reshape((num_lines,1))

            this_req_ifmap_demands = ifmap_demand_mat[start_line_idx:(end_line_idx + 1), :]
            this_req_filter_demands = filter_demand_mat[start_line_idx:(end_line_idx + 1), :]
            this_req_ofmap_demands = ofmap_demand_mat[start_line_idx:(end_line_idx + 1), :]

            no_stall_cycles = num_lines     # Since the cycles are consecutive at this point

            time_start = time.time()
            ifmap_cycles_out = \
                self.ifmap_buf.service_reads(incoming_requests_arr_np=this_req_ifmap_demands,
                                             incoming_cycles_arr=this_req_cycles_arr_np)
            time_end = time.time()
            delta = time_end - time_start
            avg_read_time_series.append(delta)

            # Take care of the incurred stalls when launching demands for filter_reads
            # Note: Stalls incurred on reading line i in ifmap reflect the request cycles for line
            #       i+1 in filter
            ifmap_hit_latency = self.ifmap_buf.get_hit_latency()
            # Vec - vec - scalar
            ifmap_stalls = ifmap_cycles_out - this_req_cycles_arr_np - ifmap_hit_latency
            # Shift by one row
            ifmap_stalls = np.concatenate((np.zeros((1,1)), ifmap_stalls[0:-1]), axis=0)
            this_req_cycles_arr_np = this_req_cycles_arr_np + ifmap_stalls

            time_start = time.time()
            filter_cycles_out = \
                self.filter_buf.service_reads(incoming_requests_arr_np=this_req_filter_demands,
                                              incoming_cycles_arr=this_req_cycles_arr_np)
            time_end = time.time()
            delta = time_end - time_start
            avg_read_time_series.append(delta)

            # Take care of stalls again --> The entire array stops when there is a stall
            filter_hit_latency = self.filter_buf.get_hit_latency()
            # Vec - vec - scalar
            filter_stalls = filter_cycles_out - this_req_cycles_arr_np - filter_hit_latency
            # Shift by one row
            filter_stalls = np.concatenate((np.zeros((1, 1)), filter_stalls[0:-1]), axis=0)
            this_req_cycles_arr_np = this_req_cycles_arr_np + filter_stalls

            ofmap_cycles_out = \
                self.ofmap_buf.service_writes(incoming_requests_arr_np=this_req_ofmap_demands,
                                              incoming_cycles_arr_np=this_req_cycles_arr_np)

            # Make the trace matrices
            this_req_ifmap_trace_matrix = \
                np.concatenate((ifmap_cycles_out, this_req_ifmap_demands), axis=1)
            this_req_filter_trace_matrix = \
                np.concatenate((filter_cycles_out, this_req_filter_demands), axis=1)
            this_req_ofmap_trace_matrix = \
                np.concatenate((ofmap_cycles_out, this_req_ofmap_demands), axis=1)

            actual_cycles = ofmap_cycles_out[-1][0] - this_req_cycles_arr_np[0][0] + 1
            num_stalls = actual_cycles - no_stall_cycles

            self.stall_cycles += num_stalls
            self.total_cycles = ofmap_cycles_out[-1][0] + 1         # OFMAP is served the last

            if first:
                first = False
                self.ifmap_trace_matrix = this_req_ifmap_trace_matrix
                self.filter_trace_matrix = this_req_filter_trace_matrix
                self.ofmap_trace_matrix = this_req_ofmap_trace_matrix
            else:
                self.ifmap_trace_matrix = \
                    np.concatenate((self.ifmap_trace_matrix, this_req_ifmap_trace_matrix), axis=0)
                self.filter_trace_matrix = \
                    np.concatenate((self.filter_trace_matrix, this_req_filter_trace_matrix), axis=0)
                self.ofmap_trace_matrix = \
                    np.concatenate((self.ofmap_trace_matrix, this_req_ofmap_trace_matrix), axis=0)

            # Update the local variable for another iteration of the while loop
            cycle_offset = ofmap_cycles_out[-1][0] + 1
            start_line_idx = end_line_idx + 1

            pbar.update(num_lines)
            # Cutoff at 0
            ofmap_lines_remaining = max(ofmap_demand_mat.shape[0] - (end_line_idx + 1), 0)
            # print("DEBUG: " + str(end_line_idx))

            if end_line_idx > ofmap_demand_mat.shape[0]:
                print('Trap')

            # if int(ofmap_lines_remaining % 1000) == 0:
            #     print("DEBUG: " + str(ofmap_lines_remaining))

            # loop_end_time = time.time()
            # loop_time = loop_end_time - loop_start_time
            # print('DEBUG: Time taken in one iteration: ' + str(loop_time))

        # At this stage there might still be some data in the active buffer of the OFMAP scratchpad
        # The following drains it and generates the OFMAP
        drain_start_cycle = self.ofmap_trace_matrix[-1][0] + 1
        self.ofmap_buf.empty_all_buffers(drain_start_cycle)

        #avg_read_time = sum(avg_read_time_series) / len(avg_read_time_series)
        #print('DEBUG: Avg time to service reads= ' + str(avg_read_time))

        pbar.close()
        # END of serving demands from memory
        self.traces_valid = True

    #
    def get_total_compute_cycles(self):
        """
        Method to get the total number of compute cycles if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return self.total_cycles

    #
    def get_stall_cycles(self):
        """
        Method to get the number of stall cycles if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return int(self.stall_cycles)

    #
    def get_ifmap_sram_start_stop_cycles(self):
        """
        Method to get the start and stop cycles of ifmap SRAM requests by the systolic array if
        trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        done = False
        for ridx in range(self.ifmap_trace_matrix.shape[0]):
            if done:
                break
            row = self.ifmap_trace_matrix[ridx,1:]
            for addr in row:
                if not addr == -1:
                    self.ifmap_sram_start_cycle = self.ifmap_trace_matrix[ridx][0]
                    done = True
                    break

        done = False
        for ridx in range(self.ifmap_trace_matrix.shape[0]):
            if done:
                break
            ridx = -1 * (ridx + 1)
            row = self.ifmap_trace_matrix[ridx,1:]
            for addr in row:
                if not addr == -1:
                    self.ifmap_sram_stop_cycle  = self.ifmap_trace_matrix[ridx][0]
                    done = True
                    break

        return self.ifmap_sram_start_cycle, self.ifmap_sram_stop_cycle

    #
    def get_filter_sram_start_stop_cycles(self):
        """
        Method to get the start and stop cycles of filter SRAM requests by the systolic array if
        trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        done = False
        for ridx in range(self.filter_trace_matrix.shape[0]):

            if done:
                break
            row = self.filter_trace_matrix[ridx, 1:]
            for addr in row:
                if not addr == -1:
                    self.filter_sram_start_cycle = self.filter_trace_matrix[ridx][0]
                    done = True
                    break

        done = False
        for ridx in range(self.filter_trace_matrix.shape[0]):
            if done:
                break
            ridx = -1 * (ridx + 1)

            row = self.filter_trace_matrix[ridx, 1:]
            for addr in row:
                if not addr == -1:
                    self.filter_sram_stop_cycle = self.filter_trace_matrix[ridx][0]
                    done = True
                    break

        return self.filter_sram_start_cycle, self.filter_sram_stop_cycle

    #
    def get_ofmap_sram_start_stop_cycles(self):
        """
        Method to get the start and stop cycles of ofmap SRAM requests by the systolic array if
        trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        done = False
        for ridx in range(self.ofmap_trace_matrix.shape[0]):
            if done:
                break
            row = self.ofmap_trace_matrix[ridx, 1:]
            for addr in row:
                if not addr == -1:
                    self.ofmap_sram_start_cycle = self.ofmap_trace_matrix[ridx][0]
                    done = True
                    break

        done = False
        for ridx in range(self.ofmap_trace_matrix.shape[0]):
            if done:
                break
            ridx = -1 * (ridx + 1)
            row = self.ofmap_trace_matrix[ridx, 1:]
            for addr in row:
                if not addr == -1:
                    self.ofmap_sram_stop_cycle = self.ofmap_trace_matrix[ridx][0]
                    done = True
                    break

        return self.ofmap_sram_start_cycle, self.ofmap_sram_stop_cycle

    #
    def get_ifmap_dram_details(self):
        """
        Method to get the start cycle, stop cycle and number of reads of DRAM requests made by the
        ifmap SRAM if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        self.ifmap_dram_reads = self.ifmap_buf.get_num_accesses()
        self.ifmap_dram_start_cycle, self.ifmap_dram_stop_cycle \
            = self.ifmap_buf.get_external_access_start_stop_cycles()

        return self.ifmap_dram_start_cycle, self.ifmap_dram_stop_cycle, self.ifmap_dram_reads

    #
    def get_filter_dram_details(self):
        """
        Method to get the start cycle, stop cycle and number of reads of DRAM requests made by the
        filter SRAM if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        self.filter_dram_reads = self.filter_buf.get_num_accesses()
        self.filter_dram_start_cycle, self.filter_dram_stop_cycle \
            = self.filter_buf.get_external_access_start_stop_cycles()

        return self.filter_dram_start_cycle, self.filter_dram_stop_cycle, self.filter_dram_reads

    #
    def get_ofmap_dram_details(self):
        """
        Method to get the start cycle, stop cycle and number of writes of DRAM requests made by the
        ofmap SRAM if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'

        self.ofmap_dram_writes = self.ofmap_buf.get_num_accesses()
        self.ofmap_dram_start_cycle, self.ofmap_dram_stop_cycle \
            = self.ofmap_buf.get_external_access_start_stop_cycles()

        return self.ofmap_dram_start_cycle, self.ofmap_dram_stop_cycle, self.ofmap_dram_writes

    #
    def get_ifmap_sram_trace_matrix(self):
        """
        Method to get the ifmap SRAM trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return self.ifmap_trace_matrix

    #
    def get_filter_sram_trace_matrix(self):
        """
        Method to get the filter SRAM trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return self.filter_trace_matrix

    #
    def get_ofmap_sram_trace_matrix(self):
        """
        Method to get the ofmap SRAM trace matrix. It contains addresses requsted by the systolic
        array and the cycles (first column) at which the requests are made.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return self.ofmap_trace_matrix

    #
    def get_sram_trace_matrices(self):
        """
        Method to get the ifmap, filter and ofmap SRAM trace matrices.
        """
        assert self.traces_valid, 'Traces not generated yet'
        return self.ifmap_trace_matrix, self.filter_trace_matrix, self.ofmap_trace_matrix

    #
    def get_ifmap_dram_trace_matrix(self):
        """
        Method to get the ifmap DRAM trace matrix. It contains addresses requsted by the ifmap SRAM
        and the cycles (first column) at which the requests are made.
        """
        return self.ifmap_buf.get_trace_matrix()

    #
    def get_filter_dram_trace_matrix(self):
        """
        Method to get the filter DRAM trace matrix. It contains addresses requsted by the filter
        SRAM and the cycles (first column) at which the requests are made.
        """
        return self.filter_buf.get_trace_matrix()

    #
    def get_ofmap_dram_trace_matrix(self):
        """
        Method to get the ofmap DRAM trace matrix. It contains addresses requsted by the ofmap SRAM
        and the cycles (first column) at which the requests are made.
        """
        return self.ofmap_buf.get_trace_matrix()

    #
    def get_dram_trace_matrices(self):
        """
        Method to get the ifmap, filter and ofmap DRAM trace matrices
        """
        dram_ifmap_trace = self.ifmap_buf.get_trace_matrix()
        dram_filter_trace = self.filter_buf.get_trace_matrix()
        dram_ofmap_trace = self.ofmap_buf.get_trace_matrix()

        return dram_ifmap_trace, dram_filter_trace, dram_ofmap_trace

    #
    def print_ifmap_sram_trace(self, filename):
        """
        Method to write the ifmap SRAM trace matrix to a file if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        np.savetxt(filename, self.ifmap_trace_matrix, fmt='%i', delimiter=",")

    #
    def print_filter_sram_trace(self, filename):
        """
        Method to write the filter SRAM trace matrix to a file if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'
        np.savetxt(filename, self.filter_trace_matrix, fmt='%i', delimiter=",")

    #
    def print_ofmap_sram_trace(self, filename):
        """
        Method to write the Ofmap SRAM trace matrix to a file if trace_valid flag is set.
        """
        assert self.traces_valid, 'Traces not generated yet'
        np.savetxt(filename, self.ofmap_trace_matrix, fmt='%i', delimiter=",")

    #
    def print_ifmap_dram_trace(self, filename):
        """
        Method to write the ifmap DRAM trace matrix to a file.
        """
        self.ifmap_buf.print_trace(filename)

    #
    def print_filter_dram_trace(self, filename):
        """
        Method to write the filter DRAM trace matrix to a file.
        """
        self.filter_buf.print_trace(filename)

    #
    def print_ofmap_dram_trace(self, filename):
        """
        Method to write the iomap DRAM trace matrix to a file.
        """
        self.ofmap_buf.print_trace(filename)



"""
This file contains the 'operand_matrix' class responsible for creating the IFMAP, Filter and OFMAP
operand matrices.
"""

import numpy as np
import math

from scalesim.topology_utils import topologies as topoutil
from scalesim.layout_utils import layouts as layoututil
from scalesim.scale_config import scale_config as cfg


class operand_matrix(object):
    """
    Class which creates the IFMAP, filter and OFMAP operand matrices to be used in compute
    simulation.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Objects from outer container classes
        self.config = cfg()
        self.topoutil = topoutil()
        self.layoututil = layoututil()

        # Layer hyper parameters
        self.layer_id = 0
        self.ifmap_rows, self.ifmap_cols = 1, 1
        self.filter_rows, self.filter_cols = 1, 1
        self.num_input_channels, self.num_filters = 1, 1
        self.row_stride, self.col_stride = 1, 1
        self.sparsity_ratio_N, self.sparsity_ratio_M = 1, 1
        self.batch_size = 1

        #  Derived hyper parameters
        self.ofmap_px_per_filt, self.conv_window_size = 1, 1
        self.ofmap_rows, self.ofmap_cols = 1, 1

        # Offsets
        self.ifmap_offset, self.filter_offset, self.ofmap_offset = 0, 10000000, 20000000
        self.matrix_offset_arr = [0, 10000000, 20000000]

        # Address matrices
        self.ifmap_addr_matrix = np.ones((self.ofmap_px_per_filt, self.conv_window_size), dtype=int)
        self.filter_addr_matrix = np.ones((self.conv_window_size, self.num_filters), dtype=int)
        self.ofmap_addr_matrix = np.ones((self.ofmap_px_per_filt, self.num_filters), dtype=int)

        # Sparsity matrices
        self.sparse_filter_array = np.ones((self.conv_window_size, self.num_filters), dtype=int)
        self.ifmap_addr_matrix_original = np.ones((self.ofmap_px_per_filt, self.conv_window_size), dtype=int)

        # Flags
        self.params_set_flag = False
        self.matrices_ready_flag = False

    #
    def set_params(self,
                   config_obj,
                   topoutil_obj,
                   layoututil_obj,
                   layer_id=0,
                   ):
        """
        Method to set the operand matrix parameters for housekeeping.
        """

        self.config = config_obj
        self.topoutil = topoutil_obj
        self.layoututil = layoututil_obj
        self.layer_id = layer_id

        # TODO: Marked for cleanup
        #my_name = 'operand_matrix.set_params(): '
        #err_prefix = 'Error: ' + my_name
        #
        # if (not len(layer_hyper_param_arr) == 7 and not len(layer_hyper_param_arr) == 8
        #        and not len(layer_hyper_param_arr)==9) or (not len(layer_calc_hyper_param_arr)==4)\
        #        or (not len(self.matrix_offset_arr) == 3):
        #    message = err_prefix + 'Invalid arguments. Exiting.'
        #    print(message)
        #    return -1

        self.ifmap_rows, self.ifmap_cols = self.topoutil.get_layer_ifmap_dims(self.layer_id)
        self.filter_rows, self.filter_cols = self.topoutil.get_layer_filter_dims(self.layer_id)
        self.num_input_channels = self.topoutil.get_layer_num_channels(self.layer_id)
        self.num_filters = self.topoutil.get_layer_num_filters(self.layer_id)
        self.row_stride, self.col_stride = self.topoutil.get_layer_strides(self.layer_id)
        self.sparsity_ratio_N, self.sparsity_ratio_M = \
            self.topoutil.get_layer_sparsity_ratio(self.layer_id)
        # TODO: Marked for cleanup
        #self.row_stride = layer_hyper_param_arr[6]
        #if len(layer_hyper_param_arr) == 8:
        #    self.col_stride = layer_hyper_param_arr[7]

        # TODO: Anand
        # TODO: Next release
        # TODO: Add an option for batching
        self.batch_size = 1

        # TODO: Marked for cleanup
        #if len(layer_hyper_param_arr) == 9:
        #    self.batch_size = layer_hyper_param_arr[8]

        # Assign the calculated hyper parameters
        self.ofmap_rows, self.ofmap_cols = self.topoutil.get_layer_ofmap_dims(self.layer_id)
        self.ofmap_rows = int(self.ofmap_rows)
        self.ofmap_cols = int(self.ofmap_cols)
        self.ofmap_px_per_filt = int(self.ofmap_rows * self.ofmap_cols)
        self.conv_window_size = int(self.topoutil.get_layer_window_size(self.layer_id))

        # Assign the offsets
        self.ifmap_offset, self.filter_offset, self.ofmap_offset \
            = self.config.get_offsets()

        # Address matrices: This is needed to take into account the updated dimensions
        self.ifmap_addr_matrix = \
            np.ones((self.ofmap_px_per_filt * self.batch_size, self.conv_window_size), dtype='>i4')
        self.filter_addr_matrix = np.ones((self.conv_window_size, self.num_filters), dtype='>i4')
        self.ofmap_addr_matrix = np.ones((self.ofmap_px_per_filt, self.num_filters), dtype='>i4')
        self.params_set_flag = True

        # TODO: This should be called from top level
        # TODO: Implement get() function for getting the matrix
        # TODO: Marked for cleanup
        # Return 0 if operand matrix generation is successful
        #self.create_operand_matrices()
        #if self.matrices_ready_flag:
        #    return True, self.ifmap_addr_matrix, self.filter_addr_matrix, self.ofmap_addr_matrix
        #else:
        #    message = err_prefix + 'Address Matrices not created. Exiting!'
        #    print(message)
        #    return False, None, None, None

    # top level function to create the operand matrices
    def create_operand_matrices(self):
        """
        Method to create IFMAP, Filter and OFMAP operand matrices.
        """
        my_name = 'operand_matrix.create_operand_matrices(): '
        err_prefix = 'Error: ' + my_name

        if not self.params_set_flag:
            message = err_prefix + 'Parameters not set yet. Run set_params(). Exiting'
            print(message)
            return -1

        retcode_1 = self.create_filter_matrix()
        retcode_2 = self.create_ifmap_matrix()
        retcode_3 = self.create_ofmap_matrix()

        retcode = retcode_1 + retcode_2 + retcode_3
        if retcode == 0:
            self.matrices_ready_flag = True

        return retcode

    # creates the ifmap operand
    def create_ifmap_matrix(self):
        """
        Method to create IFMAP operand matrix.
        """
        my_name = 'operand_matrix.create_ifmap_matrix(): '
        err_prefix = 'Error: ' + my_name

        if not self.params_set_flag:
            message = err_prefix + 'Parameters not set yet. Run set_params(). Exiting'
            print(message)
            return -1

        row_indices = np.arange(self.batch_size * self.ofmap_px_per_filt)
        col_indices = np.arange(self.conv_window_size)

        # Create 2D index arrays using meshgrid
        i, j = np.meshgrid(row_indices, col_indices, indexing='ij')

        # Call calc_ifmap_elem_addr_numpy with 2D index arrays
        self.ifmap_addr_matrix = self.calc_ifmap_elem_addr(i, j)
        self.ifmap_addr_matrix_original = self.ifmap_addr_matrix
  
        if self.config.sparsity_support:
            if self.config.sparsity_optimized_mapping is False:
                sparse_row = self.sparse_filter_array[:, 0].T
                self.ifmap_addr_matrix *= sparse_row
                non_zero_columns = np.any(self.ifmap_addr_matrix != 0, axis=0)
                if self.ifmap_addr_matrix.shape[0] == 1 and self.config.ifmap_offset == 0:
                    non_zero_columns[0] = True # Always keep the first column
                self.ifmap_addr_matrix = self.ifmap_addr_matrix[:, non_zero_columns]
  
        return 0

    # logic to translate ifmap into matrix fed into systolic array MACs
    def calc_ifmap_elem_addr(self, i, j):
        """
        Method to calculate the address of an IFMAP element.
        """
        offset = self.ifmap_offset
        ifmap_rows = self.ifmap_rows
        ifmap_cols = self.ifmap_cols
        filter_col = self.filter_cols
        r_stride = self.row_stride
        c_stride = self.col_stride
        Ew = self.ofmap_cols
        channel = self.num_input_channels

        ofmap_row, ofmap_col = np.divmod(i, Ew)
        i_row, i_col = ofmap_row * r_stride, ofmap_col * c_stride
        window_addr = (i_row * ifmap_cols + i_col) * channel

        c_row, k = np.divmod(j, filter_col * channel)
        c_col, c_ch = np.divmod(k, channel)

        valid_indices = np.logical_and(c_row + i_row < ifmap_rows, c_col + i_col < ifmap_cols)
        ifmap_px_addr = np.full(i.shape, -1)
        if valid_indices.any():
            internal_address = (c_row[valid_indices] * ifmap_cols + c_col[valid_indices]) * \
                               channel + c_ch[valid_indices]
            ifmap_px_addr[valid_indices] = internal_address + window_addr[valid_indices] + offset

        return ifmap_px_addr

    # creates the ofmap operand
    def create_ofmap_matrix(self):
        """
        Method to create OFMAP operand matrix.
        """
        my_name = 'operand_matrix.create_ofmap_matrix(): '
        err_prefix = 'Error: ' + my_name
        if not self.params_set_flag:
            message = err_prefix + 'Parameters not set yet. Run set_params(). Exiting'
            print(message)
            return -1

        row_indices = np.expand_dims(np.arange(self.ofmap_px_per_filt), axis=1)
        # if self.config.sparsity_support:
        #     _, col_indices = np.unique(np.array(self.filter_addr_matrix[0]), return_inverse=True)
        # else:
        col_indices = np.arange(self.num_filters)

        self.ofmap_addr_matrix = self.calc_ofmap_elem_addr(row_indices, col_indices)

        return 0

    # logic to translate ofmap into matrix resulting systolic array MACs
    def calc_ofmap_elem_addr(self, i, j):
        """
        Method to calculate the address of an OFMAP element.
        """
        offset = self.ofmap_offset
        num_filt = self.num_filters
        internal_address = num_filt * i + j
        ofmap_px_addr = internal_address + offset
        return ofmap_px_addr

    # creates the filter operand
    def create_filter_matrix(self):
        """
        Method to create filter operand matrix.
        """
        my_name = 'operand_matrix.create_filter_matrix(): '
        err_prefix = 'Error: ' + my_name
        if not self.params_set_flag:
            message = err_prefix + 'Parameters not set yet. Run set_params(). Exiting'
            print(message)
            return -1

        row_indices = np.expand_dims(np.arange(self.conv_window_size), axis=1)
        col_indices = np.arange(self.num_filters)

        self.filter_addr_matrix = self.calc_filter_elem_addr(row_indices, col_indices)

        if self.config.sparsity_support is True:
            if self.config.sparsity_optimized_mapping is False:
                pattern = np.concatenate([np.ones(self.sparsity_ratio_N, dtype=int),
                                        np.zeros(self.sparsity_ratio_M - self.sparsity_ratio_N,
                                                dtype=int)])
                num_repeats = (self.filter_addr_matrix.shape[0] + \
                            self.sparsity_ratio_M - 1) // self.sparsity_ratio_M
                column_values = np.tile(pattern, num_repeats)[:self.filter_addr_matrix.shape[0]]
                self.sparse_filter_array = \
                    np.tile(column_values[:, np.newaxis], (1, self.filter_addr_matrix.shape[1]))

            else:
                ratio_M = self.config.sparsity_block_size
                num_repeats = (self.filter_addr_matrix.shape[0] + ratio_M - 1) // ratio_M
                columns = []
                np.random.seed(self.config.sparsity_rand_seed)
                for col_idx in range(self.filter_addr_matrix.shape[1]):
                    ratio_N = np.random.randint(1, ratio_M // 2 + 1)
                    pattern = np.concatenate([np.ones(ratio_N, dtype=int), 
                                            np.zeros(ratio_M - ratio_N, dtype=int)])
                    column_values = np.tile(pattern, num_repeats)[:self.filter_addr_matrix.shape[0]]
                    columns.append(column_values)
                self.sparse_filter_array = np.column_stack(columns)

            self.filter_addr_matrix = np.multiply(self.filter_addr_matrix, self.sparse_filter_array)

            if self.config.sparsity_optimized_mapping is False:
                sparse_filter_matrix = []
                for col_num in range(self.filter_addr_matrix.shape[1]):
                    col_data = self.filter_addr_matrix[:, col_num]
                    condensed_col = []

                    for i in range(0, self.filter_addr_matrix.shape[0], self.sparsity_ratio_M):
                        block = col_data[i : i+self.sparsity_ratio_M]
                        block_nonzero = block[block != 0]

                        if len(block_nonzero) < self.sparsity_ratio_N:
                            padding_num = self.sparsity_ratio_N - len(block_nonzero)
                            block_nonzero = np.pad(block_nonzero, (0, padding_num), constant_values=0)
                        elif len(block_nonzero) > self.sparsity_ratio_N:
                            assert len(block_nonzero) <= self.sparsity_ratio_N, (
                                f"Excess non-zero entries ({len(block_nonzero)}) with sparsity ratio "
                                f"set to {self.sparsity_ratio_N}:{self.sparsity_ratio_M}"
                                )

                        condensed_col.extend(block_nonzero)

                    sparse_filter_matrix.append(condensed_col)

                sparse_filter_matrix = np.array(sparse_filter_matrix).T
                
                while sparse_filter_matrix.shape[0] > 0 and np.all(sparse_filter_matrix[-1] == 0):
                    sparse_filter_matrix = sparse_filter_matrix[:-1]
                self.filter_addr_matrix = sparse_filter_matrix
            else:
                remainder = self.filter_addr_matrix.shape[0] % (2 * self.config.sparsity_block_size)

                if remainder != 0:
                    padding_rows = (2 * self.config.sparsity_block_size) - remainder
                    self.filter_addr_matrix = np.vstack([
                        self.filter_addr_matrix,
                        np.zeros((padding_rows, self.filter_addr_matrix.shape[1]), dtype=int)
                    ])

                sparse_filter_matrix = np.zeros((self.filter_addr_matrix.shape[0] // 2, self.filter_addr_matrix.shape[1]), dtype=int)

                for col in range(self.filter_addr_matrix.shape[1]):
                    compressed_col = []  # To store the compressed column values
                    for start_row in range(0, self.filter_addr_matrix.shape[0], self.config.sparsity_block_size * 2):
                        block1 = self.filter_addr_matrix[start_row : start_row + self.config.sparsity_block_size, col]
                        padded_block1 = np.zeros(self.config.sparsity_block_size // 2, dtype=int)
                        non_zero_block1 = block1[block1 != 0]
                        padded_block1[:min(len(non_zero_block1), self.config.sparsity_block_size // 2)] = non_zero_block1[: self.config.sparsity_block_size // 2]

                        block2 = self.filter_addr_matrix[start_row + self.config.sparsity_block_size : start_row + self.config.sparsity_block_size * 2, col]
                        padded_block2 = np.zeros(self.config.sparsity_block_size // 2, dtype=int)
                        non_zero_block2 = block2[block2 != 0]
                        padded_block2[:min(len(non_zero_block2), self.config.sparsity_block_size // 2)] = non_zero_block2[: self.config.sparsity_block_size // 2]

                        combined_block = np.concatenate([padded_block1, padded_block2])

                        compressed_col.extend(combined_block)

                    # Update the column in the compressed matrix
                    sparse_filter_matrix[:len(compressed_col), col] = compressed_col

                # Replace the original matrix with the compressed matrix
                self.filter_addr_matrix = sparse_filter_matrix

                first_element = self.filter_addr_matrix[0][0]
                self.filter_addr_matrix[:][self.filter_addr_matrix[:] == 0] = -1
                if self.config.filter_offset == 0 and first_element == 0:
                    self.filter_addr_matrix[0][0] = 0

        return 0

    # logic to translate filter into matrix fed into systolic array MACs
    def calc_filter_elem_addr(self, i, j):
        """
        Method to calculate the address of a filter element.
        """
        offset = self.filter_offset
        filter_row = self.filter_rows
        filter_col = self.filter_cols
        channel = self.num_input_channels
        internal_address = j * filter_row * filter_col * channel + i
        filter_px_addr = internal_address + offset
        return filter_px_addr

    # function to get a part or the full ifmap operand
    def get_ifmap_matrix_part(self, start_row=0, num_rows=-1, start_col=0,
                              num_cols=-1):
        """
        Method to get a part or full IFMAP operand matrix if no error. If error, return the error
        code.
        """
        if num_rows == -1:
            num_rows = self.ofmap_px_per_filt
        if num_cols == -1:
            num_cols = self.conv_window_size
        my_name = 'operand_matrix.get_ifmap_matrix_part(): '
        err_prefix = 'Error: ' + my_name
        if not self.matrices_ready_flag:
            if self.params_set_flag:
                self.create_operand_matrices()
            else:
                message = err_prefix + ": Parameters not set yet. Run set_params(). Exiting!"
                print(message)
                return -1, np.zeros((1, 1))
        if (start_row + num_rows) > self.ofmap_px_per_filt or \
           (start_col + num_cols) > self.conv_window_size:
            message = err_prefix + ": Illegal arguments. Exiting!"
            print(message)
            return -2, np.zeros((1, 1))

        # Anand: ISSUE #3. Patch
        #end_row = start_row + num_rows + 1
        #end_col = start_col + num_cols + 1
        #ret_mat = self.ifmap_addr_matrix[start_row: end_row][start_col: end_col]
        end_row = start_row + num_rows
        end_col = start_col + num_cols
        ret_mat = self.ifmap_addr_matrix[start_row: end_row, start_col: end_col]
        return 0, ret_mat

    #
    def get_ifmap_matrix(self):
        """
        Method to get IFMAP operand matrix.
        """
        return self.get_ifmap_matrix_part()

    def get_ifmap_prefetch_matrix_custom_layout(self):
        """
            Default ifmap layout:
            for x in X:
              for y in Y:
                for c in C:
            "X->Y->C"
            Under which, the address should be calculated as

            ifmap_px_addr = internal_address + offset
        """
        #                          X, Y, C
        #ifmap_intraline_factor = [2, 2, 16] # A given fixed dimension order
        #ifmap_intraline_order  = [0, 1, 2] # A given fixed dimension order
        #ifmap_interline_order  = [1+3, 2+3, 0+3] # A given fixed dimension order
        ifmap_intraline_factor = self.layoututil.get_layer_ifmap_intraline_factor()
        ifmap_intraline_order = self.layoututil.get_layer_ifmap_intraline_order()
        ifmap_interline_order = self.layoututil.get_layer_ifmap_interline_order()
        
        # Sanity Checking
        assert np.prod(ifmap_intraline_factor) <= int(self.config.get_ifmap_sram_bandwidth())

        ifmap_overall_data = np.arange(self.ifmap_rows * self.ifmap_cols * self.num_input_channels)
        
        # Reshape each dimension of ifmap to be a different index.
        ifmap_overall_data = ifmap_overall_data.reshape(self.ifmap_rows, self.ifmap_cols, self.num_input_channels)
        num_x_t = math.ceil(self.ifmap_rows / ifmap_intraline_factor[0])
        padding_x = num_x_t * ifmap_intraline_factor[0] - self.ifmap_rows
        num_y_t = math.ceil(self.ifmap_cols / ifmap_intraline_factor[1])
        padding_y = num_y_t * ifmap_intraline_factor[1] - self.ifmap_cols
        num_c_t = math.ceil(self.num_input_channels / ifmap_intraline_factor[2])
        padding_c = num_c_t * ifmap_intraline_factor[2] - self.num_input_channels
        x_t = ifmap_intraline_factor[0]
        y_t = ifmap_intraline_factor[1]
        c_t = ifmap_intraline_factor[2]

        ifmap_overall_data_pad = np.pad(ifmap_overall_data, [(0, padding_x), (0, padding_y), (0, padding_c)], 'constant', constant_values=-1)
        # Note that: padded -1 will be removed in read_buffer/prepare_hashed_buffer, so won't waste on-chip memory.
        
        # Factorize each dimension of ifmap to be interline (row) and intraline (col).
        assert ifmap_overall_data_pad.shape[0] % ifmap_intraline_factor[0] == 0, (f"dimension X must be full divisible by"
             f" intraline factorization; post-padding X:{ifmap_overall_data_pad.shape[0]}, factor:{ifmap_intraline_factor[0]}")
        assert ifmap_overall_data_pad.shape[1] % ifmap_intraline_factor[1] == 0, (f"dimension Y must be full divisible by"
             f" intraline factorization; post-padding Y:{ifmap_overall_data_pad.shape[1]} factor:{ifmap_intraline_factor[1]}")
        assert ifmap_overall_data_pad.shape[2] % ifmap_intraline_factor[2] == 0, (f"dimension C must be full divisible by" 
             f" intraline factorization; post-padding C:{ifmap_overall_data_pad.shape[2]}, factor:{ifmap_intraline_factor[2]}")
        
        """
            x_t = ifmap_intraline_factor[1]
            y_t = ifmap_intraline_factor[2]
            c_t = ifmap_intraline_factor[3]
            
            for x_1 in range(X, x_t):
              for x_2 in range(x_t):
                for y_1 in range(Y, y1):
                  for y_2 in range(y1):
                    for c_1 in range(C, c_t):
                      for c_2 in range(c_t):
                        # Address specification.
                        # x = x_2 + x_1 * x_t
                        # y = y_2 + y_1 * y_t
                        # c = c_2 + c_1 * c1
        """
        ifmap_overall_data_pad = ifmap_overall_data_pad.reshape(num_x_t, x_t, num_y_t, y_t, num_c_t, c_t)
        
        # Transpose two separate interline dimension order and intraline dimension order.
        ifmap_overall_data_pad = np.transpose(ifmap_overall_data_pad, (0,2,4,1,3,5))

        # Transpose data layout to force a specific order.
        ifmap_overall_data_pad = np.transpose(ifmap_overall_data_pad, 
           (ifmap_interline_order[0], ifmap_interline_order[1], ifmap_interline_order[2], 
            ifmap_intraline_order[0], ifmap_intraline_order[1], ifmap_intraline_order[2]))

        return ifmap_overall_data_pad.reshape(1,-1)
        
    # function to get a part or the full filter operand
    def get_filter_matrix_part(self, start_row=0, num_rows=-1, start_col=0,
                               num_cols=-1):
        """
        Method to get a part or full filter operand matrix if no error. If error, return the error
        code.
        """

        if num_rows == -1:
            num_rows = self.conv_window_size
        if num_cols == -1:
            num_cols = self.num_filters

        my_name = 'operand_matrix.get_filter_matrix_part(): '
        err_prefix = 'Error: ' + my_name
        if not self.matrices_ready_flag:
            if self.params_set_flag:
                self.create_operand_matrices()
            else:
                message = err_prefix + ": Parameters not set yet. Run set_params(). Exiting!"
                print(message)
                return -1, np.zeros((1, 1))
        if (start_row + num_rows) > self.conv_window_size or \
           (start_col + num_cols) > self.num_filters:
            message = err_prefix + ": Illegal arguments. Exiting!"
            print(message)
            return -2, np.zeros((1, 1))

        # Anand: ISSUE #3. FIX
        #end_row = start_row + num_rows + 1
        #end_col = start_col + num_cols + 1
        end_row = start_row + num_rows
        end_col = start_col + num_cols

        # Anand: ISSUE #3. FIX
        #ret_mat = self.filter_addr_matrix[start_row: end_row][start_col: end_col]
        # ret_mat = self.filter_addr_matrix[start_row: end_row, start_col: end_col]
        if self.config.sparsity_support is True:
            ret_mat = self.filter_addr_matrix
        else:
            ret_mat = self.filter_addr_matrix[start_row: end_row, start_col: end_col]

        return 0, ret_mat

    #
    def get_filter_matrix(self):
        """
        Method to get filter operand matrix.
        """
        return self.get_filter_matrix_part()


    def get_filter_prefetch_matrix_custom_layout(self):
        """
            Default filter layout:
            for k in K:
              for c in C:
                for r in R:
                  for s in S:
            "K->C->R->S"
            Under which, the address should be calculated as

            internal_address = j * filter_row * filter_col * channel + i
            j:  the kernel index 
        """
        #                            K, C, R, S
        # filter_intraline_factor = [4,16, 1, 1] # A given fixed dimension order
        # filter_intraline_order  = [3, 2, 1, 0] # A given fixed dimension order
        # filter_interline_order  = [0+4, 1+4, 2+4, 3+4] # A given fixed dimension order
        filter_intraline_factor = self.layoututil.get_layer_filter_intraline_factor()
        filter_intraline_order  = self.layoututil.get_layer_filter_intraline_order()
        filter_interline_order  = self.layoututil.get_layer_filter_interline_order()

        # Sanity Checking
        assert np.prod(filter_intraline_factor) == int(self.config.get_filter_sram_bandwidth())

        filter_overall_data = np.arange(self.filter_rows * self.filter_cols * self.num_input_channels * self.num_filters) + self.filter_offset
        
        # Reshape each dimension of filter to be a different index.
        filter_overall_data = filter_overall_data.reshape(self.num_filters, self.num_input_channels, self.filter_rows, self.filter_cols)
        num_k_t = math.ceil(self.num_filters / filter_intraline_factor[0])
        padding_k = num_k_t * filter_intraline_factor[0] - self.num_filters
        num_c_t = math.ceil(self.num_input_channels / filter_intraline_factor[1])
        padding_c = num_c_t * filter_intraline_factor[1] - self.num_input_channels
        num_r_t = math.ceil(self.filter_rows / filter_intraline_factor[2])
        padding_r = num_r_t * filter_intraline_factor[2] - self.filter_rows
        num_s_t = math.ceil(self.filter_cols / filter_intraline_factor[3])
        padding_s = num_s_t * filter_intraline_factor[3] - self.filter_cols
        k_t = filter_intraline_factor[0]
        c_t = filter_intraline_factor[1]
        r_t = filter_intraline_factor[2]
        s_t = filter_intraline_factor[3]

        filter_overall_data_pad = np.pad(filter_overall_data, [(0, padding_k), (0, padding_c), (0, padding_r), (0, padding_s)], 'constant', constant_values=-1)
        # Note that: padded -1 will be removed in read_buffer/prepare_hashed_buffer, so won't waste on-chip memory.
        
        # Factorize each dimension of filter to be interline (row) and intraline (col).
        assert filter_overall_data_pad.shape[0] % filter_intraline_factor[0] == 0, (f"dimension K must be full divisible by"
             f" intraline factorization; post-padding K:{filter_overall_data_pad.shape[0]}, factor:{filter_intraline_factor[0]}")
        assert filter_overall_data_pad.shape[1] % filter_intraline_factor[1] == 0, (f"dimension C must be full divisible by"
             f" intraline factorization; post-padding C:{filter_overall_data_pad.shape[1]} factor:{filter_intraline_factor[1]}")
        assert filter_overall_data_pad.shape[2] % filter_intraline_factor[2] == 0, (f"dimension R must be full divisible by" 
             f" intraline factorization; post-padding R:{filter_overall_data_pad.shape[2]}, factor:{filter_intraline_factor[2]}")
        assert filter_overall_data_pad.shape[3] % filter_intraline_factor[3] == 0, (f"dimension S must be full divisible by" 
             f" intraline factorization; post-padding S:{filter_overall_data_pad.shape[2]}, factor:{filter_intraline_factor[2]}")
        
        """
            k_t = filter_intraline_factor[0]
            c_t = filter_intraline_factor[1]
            r_t = filter_intraline_factor[2]
            s_t = filter_intraline_factor[3]

            for k_1 in range(K, k_t):
              for k_2 in range(k_t):
                for c_1 in range(C, c_t):
                  for c_2 in range(c_t):
                    for r_1 in range(R, r_t):
                      for r_2 in range(r_t):
                        for s_1 in range(S, s_t):
                          for s_2 in range(s_t):
                            # Address specification.
                            # k = k_2 + k_1 * k_t
                            # c = c_2 + c_1 * c_t
                            # r = r_2 + r_1 * r1
                            # s = s_2 + s_1 * s1
        """
        filter_overall_data_pad = filter_overall_data_pad.reshape(num_k_t, k_t, num_c_t, c_t, num_r_t, r_t, num_s_t, s_t)
        
        # Transpose two separate interline dimension order and intraline dimension order.
        filter_overall_data_pad = np.transpose(filter_overall_data_pad, (0,2,4,6,1,3,5,7))

        # Transpose data layout to force a specific order.
        filter_overall_data_pad = np.transpose(filter_overall_data_pad, 
           (filter_interline_order[0], filter_interline_order[1], filter_interline_order[2], filter_interline_order[3], 
            filter_intraline_order[0], filter_intraline_order[1], filter_intraline_order[2], filter_intraline_order[3]))

        print(f"finalized filter.shape = {filter_overall_data_pad.shape}")
        return filter_overall_data_pad.reshape(1,-1)
    
    # function to get a part or the full ofmap operand
    def get_ofmap_matrix_part(self, start_row=0, num_rows=-1, start_col=0,
                               num_cols=-1):
        """
        Method to get a part or full OFMAP operand matrix if no error. If error, return the error
        code.
        """

        # Since we cannot pass self as an argument in the member functions
        # This is an alternate way of making the matrix dimensions as defaults
        if num_rows == -1:
            num_rows = self.ofmap_px_per_filt
        if num_cols == -1:
            num_cols = self.num_filters
        my_name = 'operand_matrix.get_ofmap_matrix_part(): '
        err_prefix = 'Error: ' + my_name
        if not self.matrices_ready_flag:
            if self.params_set_flag:
                self.create_operand_matrices()
            else:
                message = err_prefix + ": Parameters not set yet. Run set_params(). Exiting!"
                print(message)
                return -1, np.zeros((1, 1))
        if (start_row + num_rows) > self.ofmap_px_per_filt or \
           (start_col + num_cols) > self.num_filters:
            message = err_prefix + ": Illegal arguments. Exiting!"
            print(message)
            return -2, np.zeros((1, 1))

        # Anand: ISSUE #3. Patch
        #end_row = start_row + num_rows + 1
        #end_col = start_col + num_cols + 1
        #ret_mat = self.filter_addr_matrix[start_row: end_row][start_col: end_col]
        end_row = start_row + num_rows
        end_col = start_col + num_cols
        # Anand: ISSUE #7. Patch
        #ret_mat = self.filter_addr_matrix[start_row: end_row, start_col: end_col]
        if self.config.sparsity_support is True:
            ret_mat =self.ofmap_addr_matrix
        else:
            ret_mat = self.ofmap_addr_matrix[start_row: end_row, start_col: end_col]

        return 0, ret_mat

    #
    def get_ofmap_matrix(self):
        """
        Method to get OFMAP operand matrix.
        """
        return self.get_ofmap_matrix_part()

    #
    def get_all_operand_matrix(self):
        """
        Method to get IFMAP, Filter and OFMAP operand matrices.
        """
        if not self.matrices_ready_flag:
            me = 'operand_matrix.' + 'get_all_operand_matrix()'
            message = 'ERROR:' + me + ': Matrices not ready or matrix gen failed'
            print(message)
            return

        return self.ifmap_addr_matrix, \
               self.filter_addr_matrix, \
               self.ofmap_addr_matrix


if __name__ == '__main__':
    # opmat = operand_matrix()
    tutil = topoutil()
    layout = layoututil()
    lid = 3
    topology_file = "../../topologies/conv_nets/test.csv"
    layout_file = "../../layouts/conv_nets/test.csv"
    tutil.load_arrays(topofile=topology_file)
    layout.load_arrays(layoutfile=layout_file)
    for i in range(tutil.get_num_layers()):
        layer_param_arr = tutil.get_layer_params(layer_id=i)
        ofmap_dims = tutil.get_layer_ofmap_dims(layer_id=i)
        ofmap_px_filt = tutil.get_layer_num_ofmap_px(layer_id=i) / tutil.get_layer_num_filters(layer_id=i)
        conv_window_size = tutil.get_layer_window_size(layer_id=i)

        layer_calc_hyper_param_arr = [ofmap_dims[0], ofmap_dims[1], ofmap_px_filt, conv_window_size]
        config_arr = [512, 512, 256, 8, 8]
        #[matrix_set, ifmap_addr_matrix, filter_addr_matrix, ofmap_addr_matrix] \
        #    = opmat.set_params(layer_hyper_param_arr=layer_param_arr[1:],
        #                       layer_calc_hyper_param_arr=layer_calc_hyper_param_arr,
        #                       offset_list=[0, 1000000, 2000000])

        ifmap_intraline_factor = layout.get_layer_ifmap_intraline_factor(layer_id=i)
        ifmap_intraline_order = layout.get_layer_ifmap_intraline_order(layer_id=i)
        ifmap_interline_order = layout.get_layer_ifmap_interline_order(layer_id=i)

        filter_intraline_factor = layout.get_layer_filter_intraline_factor(layer_id=i)
        filter_intraline_order = layout.get_layer_filter_intraline_order(layer_id=i)
        filter_interline_order = layout.get_layer_filter_interline_order(layer_id=i)

        print(f"ifmap_intraline_factor={ifmap_intraline_factor}")
        print(f"ifmap_intraline_order={ifmap_intraline_order}")
        print(f"ifmap_interline_order={ifmap_interline_order}")
        print(f"filter_intraline_factor={filter_intraline_factor}")
        print(f"filter_intraline_order={filter_intraline_order}")
        print(f"filter_interline_order={filter_interline_order}")



"""
The 'ReadBufferEstimateBw' class simulates a double-buffered memory read buffer in estimate
bandwidth mode designed for systolic array computations.
"""

import math
import numpy as np

from scalesim.memory.read_port import read_port


class ReadBufferEstimateBw:
    """
    Class which service the estimate bandwidth mode in the read buffer.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Buffer parameters
        self.word_size = 1
        self.active_buf_frac = 0.5
        self.total_size_bytes = 1
        self.total_size_elems = 1
        self.active_buf_size = 1
        self.prefetch_buf_size = 1

        self.hit_latency = 1

        # Backing buffer parameters
        self.backing_buffer = read_port()
        self.default_bandwidth = 1
        self.prefetch_bandwidth = 1

        # Access counts
        self.num_access = 0

        # Trace matrix
        self.trace_matrix = np.ones((1, 1))

        # Tracking variables
        self.num_items_per_set = -1
        self.elems_current_set = 0
        self.current_set_id = 0
        self.read_buffer_set_start_id = -1
        self.read_buffer_set_end_id = -1
        self.prefetch_buffer_set_start_id = -1
        self.prefetch_buffer_set_end_id = -1
        self.last_prefetch_start_cycle = -2
        self.last_prefetch_end_cycle = -1
        self.first_request_rcvd_cycle = 0

        # Internal data structures
        self.current_set = set()
        self.list_of_sets = []
        self.num_sets_active_buffer = 1
        self.num_sets_prefetch_buffer = 1

        # Flags
        self.first_request_seen = False
        self.params_set_flag = False
        self.active_buffer_prefetch_done = False
        self.trace_valid = False
        self.use_ramulator_trace = False

    #
    def set_params(self, backing_buf_obj,
                   total_size_bytes=1, word_size=1, active_buf_frac=0.9,
                   hit_latency=1, backing_buf_default_bw=1,
                   use_ramulator_trace = False):
        """
        Method to set the ifmap/filter double buffered memory simulation parameters for estimate
        bandwidth mode.
        """

        self.total_size_bytes = total_size_bytes
        self.word_size = word_size

        assert 0.5 <= active_buf_frac < 1, "Valid active buf frac [0.5,1)"
        self.active_buf_frac = round(active_buf_frac, 2)
        self.hit_latency = hit_latency

        self.backing_buffer = backing_buf_obj
        self.default_bandwidth = backing_buf_default_bw
        self.prefetch_bandwidth = self.default_bandwidth

        # Calculate these based on the values provided
        self.total_size_elems = math.floor(self.total_size_bytes / self.word_size)
        self.active_buf_size = int(math.ceil(self.total_size_elems * self.active_buf_frac))
        self.prefetch_buf_size = self.total_size_elems - self.active_buf_size

        #
        self.num_items_per_set = math.floor(self.total_size_elems / 100)
        self.num_sets_active_buffer = int(self.active_buf_frac * 100)
        self.num_sets_prefetch_buffer = 100 - self.num_sets_active_buffer

        self.current_set = set()
        self.current_set_id = 0
        self.list_of_sets = []
        self.read_buffer_set_start_id = 0
        self.read_buffer_set_end_id = self.num_sets_active_buffer - 1
        self.last_prefetch_start_cycle = -2
        self.last_prefetch_end_cycle = -1  # TODO: Check what the correct value is
        self.use_ramulator_trace = use_ramulator_trace

        #
        self.params_set_flag = True

    #
    def service_reads(self, incoming_requests_arr_np, incoming_cycles_arr):
        """
        Method to service read requests coming from systolic array in estimate bandwidth mode.
        """
        assert self.params_set_flag, 'Parameters are not set yet'
        assert incoming_cycles_arr.shape[0] == incoming_requests_arr_np.shape[0],\
               'Incoming cycles and requests dont match'

        outcycles = incoming_cycles_arr + self.hit_latency
        # In estimate mode, operation is stall free.
        # Therefore its always a hit

        # The following to track requests and maintain proper state of the buffer
        for i in range(incoming_requests_arr_np.shape[0]):
            cycle = int(incoming_cycles_arr[i][0])

            requests_this_cycle = incoming_requests_arr_np[i]
            if not self.first_request_seen:
                if max(requests_this_cycle) > -1:
                    self.first_request_rcvd_cycle = cycle
                    self.first_request_seen = True

            for addr in requests_this_cycle:
                if not addr == -1:
                    self.manage_prefetches(cycle, addr)

        return outcycles

    #
    def manage_prefetches(self, cycle, addr):
        """
        Method to manage prefetches in estimate bandwidth mode.
        """

        # If this is a new address, otherwise its a hit
        if self.check_hit(addr):
            return

        if addr not in self.current_set:
            self.current_set.add(addr)
            self.elems_current_set += 1

            if self.elems_current_set == self.num_items_per_set:
                self.list_of_sets += [self.current_set]
                self.current_set = set()
                self.elems_current_set = 0
                self.current_set_id += 1

                # This should be prefetched
                if self.current_set_id == self.read_buffer_set_end_id + 1:
                    if not self.active_buffer_prefetch_done:
                        self.prefetch_bandwidth = self.default_bandwidth
                        self.last_prefetch_end_cycle = \
                            self.first_request_rcvd_cycle - 1 - self.backing_buffer.get_latency()

                        cycles_needed = (self.num_sets_prefetch_buffer * self.num_items_per_set) \
                                        / self.prefetch_bandwidth
                        cycles_needed = math.ceil(cycles_needed)

                        self.last_prefetch_start_cycle = \
                            self.last_prefetch_end_cycle - cycles_needed + 1

                        self.prefetch()
                        self.prefetch_buffer_set_start_id =self.read_buffer_set_end_id + 1
                        self.prefetch_buffer_set_end_id = self.prefetch_buffer_set_start_id + \
                                                          self.num_sets_prefetch_buffer - 1
                        self.active_buffer_prefetch_done = True

                    else:
                        elems_to_prefetch = self.num_sets_prefetch_buffer * self.num_items_per_set
                        cycles_needed = \
                            self.last_prefetch_end_cycle - self.last_prefetch_start_cycle + 1
                        self.prefetch_bandwidth = math.ceil(elems_to_prefetch / cycles_needed)
                        self.prefetch()
                        self.prefetch_buffer_set_start_id += self.num_sets_prefetch_buffer
                        self.prefetch_buffer_set_end_id += self.num_sets_prefetch_buffer

                    # Solving memory leak by discarding sets that are no longer in use
                    i = self.read_buffer_set_start_id
                    for j in range(self.num_sets_prefetch_buffer):
                        self.list_of_sets[i+j] = None


                    self.read_buffer_set_start_id += self.num_sets_prefetch_buffer
                    self.read_buffer_set_end_id += self.num_sets_prefetch_buffer
                    self.last_prefetch_start_cycle = self.last_prefetch_end_cycle +1
                    self.last_prefetch_end_cycle = cycle

    #
    def check_hit(self, addr):
        """
        Method to check if the address is hit or miss in the active read buffer.
        """
        assert self.params_set_flag, 'Parameters are not set yet'

        start_set_idx = self.read_buffer_set_start_id
        end_set_idx = min(self.current_set_id, self.read_buffer_set_end_id + 1)

        if start_set_idx == end_set_idx:
            return False

        for idx in range(start_set_idx, end_set_idx):
            if addr in self.list_of_sets[idx]:
                return True

        return False

    #
    def complete_all_prefetches(self):
        """
        Method to complete all the prefetches in estimate bandwidth mode. Prefetch first the active
        buffer if not done before and then keep prefetching prefetch buffers.
        """
        assert self.params_set_flag, 'Parameters are not set yet'

        current_set_elems = list(self.current_set)
        if len(current_set_elems) > 0:
            self.list_of_sets += [self.current_set]
        else:
            self.current_set_id -= 1    # If there are no elems in this set, dont consider it

        if not self.active_buffer_prefetch_done:
            self.prefetch_bandwidth = self.default_bandwidth
            self.last_prefetch_end_cycle = -1 - self.backing_buffer.get_latency()

            num_sets_to_prefetch = self.current_set_id + 1
            self.num_sets_active_buffer = num_sets_to_prefetch

            cycles_needed = (num_sets_to_prefetch * self.num_items_per_set) \
                            / self.prefetch_bandwidth
            cycles_needed = math.ceil(cycles_needed)

            self.last_prefetch_start_cycle = self.last_prefetch_end_cycle - cycles_needed + 1

            self.prefetch()
            self.active_buffer_prefetch_done = True
        else:
            num_sets_to_prefetch = self.current_set_id - self.prefetch_buffer_set_start_id + 1
            self.prefetch_buffer_set_end_id = self.current_set_id
            elems_to_prefetch = num_sets_to_prefetch * self.num_items_per_set
            cycles_needed = self.last_prefetch_end_cycle - self.last_prefetch_start_cycle + 1
            self.prefetch_bandwidth = math.ceil(elems_to_prefetch / cycles_needed)
            self.prefetch()

    #
    def prefetch(self):
        """
        Method to do a new prefetch. In a new prefetch, some portion of the original data needs to
        be deleted to accomodate the prefetched data In this case we overwrite some data in the
        active buffer with the prefetched data and then create a new prefetch request.
        """
        assert self.params_set_flag, 'Parameters are not set yet'

        if not self.active_buffer_prefetch_done:
            start_set_idx = 0
            end_set_idx = self.num_sets_active_buffer - 1
        else:
            start_set_idx = self.prefetch_buffer_set_start_id
            end_set_idx = self.prefetch_buffer_set_end_id

        all_addresses = []
        for idx in range(start_set_idx, end_set_idx + 1):
            this_set = self.list_of_sets[idx]
            all_addresses += list(this_set)

        self.num_access += len(all_addresses)

        cycles_needed = self.last_prefetch_end_cycle - self.last_prefetch_start_cycle + 1
        max_prefetch_capacity = cycles_needed * self.prefetch_bandwidth

        delta = max_prefetch_capacity - len(all_addresses)

        if delta > 0:
            for _ in range(delta):
                all_addresses += [-1]

        prefetch_requests = np.asarray(all_addresses).reshape((cycles_needed,
                                                               self.prefetch_bandwidth))

        cycles_arr = np.zeros((cycles_needed,1))
        for i in range(cycles_arr.shape[0]):
            cycles_arr[i][0] = self.last_prefetch_start_cycle + i

        response_cycles_arr = \
            self.backing_buffer.service_reads(incoming_cycles_arr=cycles_arr,
                                              incoming_requests_arr_np=prefetch_requests)

        # Create / add elements to the trace matrix
        this_prefetch_traces = np.concatenate((response_cycles_arr, prefetch_requests), axis=1)

        if not self.trace_valid:
            self.trace_matrix = this_prefetch_traces
            self.trace_valid = True

        else:
            del_cols = self.trace_matrix.shape[1] - this_prefetch_traces.shape[1]
            if del_cols > 0:
                empty_cols = np.ones((this_prefetch_traces.shape[0], del_cols))
                this_prefetch_traces = np.concatenate((this_prefetch_traces, empty_cols), axis=1)

            elif del_cols < 0:
                del_cols = int(-1 * del_cols)
                empty_cols = np.ones((self.trace_matrix.shape[0], del_cols))
                self.trace_matrix = np.concatenate((self.trace_matrix, empty_cols), axis=1)

            self.trace_matrix = np.concatenate((self.trace_matrix, this_prefetch_traces), axis=0)

    #
    def get_latency(self):
        """
        Method to get hit latency of the read estimate buffer.
        """
        assert self.params_set_flag, 'Parameters are not valid'
        return self.hit_latency

    #
    def get_trace_matrix(self):
        """
        Method to get the read estimate buffer trace matrix. It contains addresses requsted by the
        systolic array and the cycles (first column) at which the requests are made.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        return self.trace_matrix

    #
    def get_hit_latency(self):
        """
        Method to get hit latency of the read estimate buffer.
        """
        return self.hit_latency

    #
    def get_num_accesses(self):
        """
        Method to get number of accesses of the read estimate buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        return self.num_access

    #
    def get_external_access_start_stop_cycles(self):
        """
        Method to get start and stop cycles of the read estimate buffer if trace_valid flag is set.
        """
        assert self.trace_valid, 'Traces not ready yet'
        start_cycle = self.trace_matrix[0][0]
        end_cycle = self.trace_matrix[-1][0]

        return start_cycle, end_cycle

    #
    def print_trace(self, filename):
        """
        Method to write the read estimate buffer trace matrix to a file.
        """
        if not self.trace_valid:
            print('No trace has been generated yet')
            return

        np.savetxt(filename, self.trace_matrix, fmt='%s', delimiter=",")



import math


class layouts(object):

    def __init__(self):
        self.current_layout_name = ""
        self.layout_file_name = ""
        self.layout_arrays = []
        self.spatio_temp_dim_arrays = []
        self.layers_calculated_hyperparams = []
        self.num_layers = 0
        self.layout_load_flag = False
        self.layout_calc_hyper_param_flag = False
        self.layout_calc_spatiotemp_params_flag = False

    # reset layoutlogy parameters
    def reset(self):
        print("All data reset")
        self.current_layout_name = ""
        self.layout_file_name = ""
        self.layout_load_flag = False
        self.layout_arrays = []
        self.num_layers = 0
        self.layout_calc_hyper_param_flag = False
        self.layers_calculated_hyperparams = []

    #
    def load_layer_params_from_list(self, layer_name, elems_list=[]):
        self.layout_file_name = ''
        self.current_layoutname = ''
        self.layer_name = layer_name
        self.append_layout_arrays(layer_name, elems_list)

        self.num_layers += 1
        self.layout_load_flag = True

    #
    def load_arrays(self, layoutfile='', mnk_inputs=False):
        self.load_layout_conv(layoutfile)

    def load_layout_conv(self, layoutfile):
        first = True
        self.layout_file_name = layoutfile.split('/')[-1]
        name_arr = self.layout_file_name.split('.')
        if len(name_arr) > 1:
            self.current_layout_name = self.layout_file_name.split('.')[-2]
        else:
            self.current_layout_name = self.layout_file_name
        f = open(layoutfile, 'r')
        for row in f:
            row = row.strip()
            if first or row == '':
                first = False
            else:
                elems = row.split(',')[:-1]
                # depth-wise convolution
                if 'DP' in elems[0].strip():
                    for dp_layer in range(int(elems[5].strip())):
                        layer_name = elems[0].strip() + "Channel_" + str(dp_layer)
                        elems[5] = str(1)
                        self.append_layout_arrays(layer_name, elems)
                else:
                    layer_name = elems[0].strip()
                    self.append_layout_arrays(layer_name, elems)

        self.num_layers = len(self.layout_arrays)
        self.layout_load_flag = True


    # Write the contents into a csv file
    def write_layout_file(self,
                      path="",
                      filename=""
                      ):
        if path == "":
            print("WARNING: layoutlogy_utils.write_layout_file: No path specified writing to the cwd")
            path = "./" 

        if filename == "":
            print("ERROR: layoutlogy_utils.write_layout_file: No filename provided")
            return

        filename = path + "/" + filename

        if not self.layout_load_flag:
            print("ERROR: layoutlogy_utils.write_layout_file: No data loaded")
            return

        header = [
                    "Layer name",
                    "IFMAP Height Intraline Factor",
                    "IFMAP Width Intraline Factor",
                    "Filter Height Intraline Factor",
                    "Filter Width Intraline Factor",
                    "Channel Intraline Factor",
                    "Num Filter Intraline Factor",
                    "IFMAP Height Intraline Order",
                    "IFMAP Width Intraline Order",
                    "Channel Intraline Order",
                    "IFMAP Height Interline Order",
                    "IFMAP Width Interline Order",
                    "Channel Interline Order",
                    "Num Filter Intraline Order",
                    "Channel Intraline Order",
                    "Filter Height Intraline Order",
                    "Filter Width Intraline Order",
                    "Num Filter Interline Order",
                    "Channel Interline Order",
                    "Filter Height Interline Order",
                    "Filter Width Interline Order"
                ]

        f = open(filename, 'w')
        log = ",".join(header)
        log += ",\n"
        f.write(log)

        for param_arr in self.layout_arrays:
            log = ",".join([str(x) for x in param_arr])
            log += ",\n"
            f.write(log)

        f.close()


    def append_layout_arrays(self, layer_name, elems):
        entry = [layer_name]

        for i in range(1, len(elems)):
            val = int(str(elems[i]).strip())
            entry.append(val)
            if i == 7 and len(elems) < 9:
                entry.append(val)  # Add the same stride in the col direction automatically

        self.layout_arrays.append(entry)


    def get_current_layout_name(self):
        current_layout_name = ""
        if self.layout_load_flag:
            current_layout_name = self.current_layout_name
        else:
            print('Error: get_current_layout_name(): layout file not read')
        return current_layout_name

    def get_num_layers(self):
        if not self.layout_load_flag:
            print("ERROR: layout.get_num_layers: No array loaded")
            return
        return self.num_layers

    #
    def get_layer_ifmap_intraline_factor(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_ifmap_intraline_factor: Invalid layer id")

        return [self.layout_arrays[layer_id][1], self.layout_arrays[layer_id][2], self.layout_arrays[layer_id][5]]

    #
    def get_layer_ifmap_intraline_order(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_ifmap_intraline_order: Invalid layer id")

        return self.layout_arrays[layer_id][7:10]

    #
    def get_layer_ifmap_interline_order(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_ifmap_interline_order: Invalid layer id")

        return self.layout_arrays[layer_id][10:13]

    #
    def get_layer_filter_intraline_factor(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_filter_intraline_factor: Invalid layer id")

        return [self.layout_arrays[layer_id][6], self.layout_arrays[layer_id][5], self.layout_arrays[layer_id][3], self.layout_arrays[layer_id][4]]

    #
    def get_layer_filter_intraline_order(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_filter_intraline_order: Invalid layer id")

        return self.layout_arrays[layer_id][13:17]

    #
    def get_layer_filter_interline_order(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_filter_interline_order: Invalid layer id")

        return self.layout_arrays[layer_id][17:]


    def get_layer_params(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_params: Invalid layer id")
            return
        layer_params = self.layout_arrays[layer_id]
        return layer_params

    def get_layer_id_from_name(self, layer_name=""):
        if (not self.layout_load_flag) or layer_name == "":
            print("ERROR")
            return
        indx = -1
        for i in range(len(self.layout_arrays)):
            if layer_name == self.layout_arrays[i]:
                indx = i
        if indx == -1:
            print("WARNING: Not found")
        return indx

    #
    def get_layer_name(self, layer_id=0):
        if not (self.layout_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: layout.get_layer_name: Invalid layer id")
            return

        name = self.layout_arrays[layer_id][0]
        return str(name)

    #
    def get_layer_names(self):
        if not self.layout_load_flag:
            print("ERROR")
            return
        layer_names = []
        for entry in self.layout_arrays:
            layer_name = str(entry[0])
            layer_names.append(layer_name)
        return layer_names


if __name__ == '__main__':
    import os
    layout = layouts()
    layout.load_arrays(os.path.join(os.getcwd(), "../layouts/GEMM_mnk/vit_s_downscale_KM_KN.csv"))
    for i in range(layout.get_num_layers()):
        ifmap_intraline_factor = layout.get_layer_ifmap_intraline_factor(layer_id=i)
        ifmap_intraline_order = layout.get_layer_ifmap_intraline_order(layer_id=i)
        ifmap_interline_order = layout.get_layer_ifmap_interline_order(layer_id=i)

        filter_intraline_factor = layout.get_layer_filter_intraline_factor(layer_id=i)
        filter_intraline_order = layout.get_layer_filter_intraline_order(layer_id=i)
        filter_interline_order = layout.get_layer_filter_interline_order(layer_id=i)

        print(f"########### Layer {i} ###########")
        print(f"ifmap_intraline_factor={ifmap_intraline_factor}")
        print(f"ifmap_intraline_order={ifmap_intraline_order}")
        print(f"ifmap_interline_order={ifmap_interline_order}")
        print(f"filter_intraline_factor={filter_intraline_factor}")
        print(f"filter_intraline_order={filter_intraline_order}")
        print(f"filter_interline_order={filter_interline_order}")

        print(f"layout.layout_arrays={layout.layout_arrays}")



"""
External DRAM read requests serviced by Ramulator
"""
import numpy as np
from scalesim.scale_config import scale_config as config
from bisect import bisect_left

class read_port:
    """
    Class to define external DRAM memory requests with Ramulator
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        self.latency = 1
        self.ramulator_trace = False
        self.latency_matrix = []
        self.bw = 10
        self.request_queue_size = 100
        self.request_queue_status = 0
        self.stall_cycles = 0
        self.request_array = []
        self.count = 0
        self.config = config()
    #
    def def_params( self,
                    config = config(),
                    latency_file = ''
                ):
        """
        Method to define the paths of ramulator trace numpy files 
        and read request queue sizes.
        """
        self.config = config
        self.ramulator_trace = self.config.get_ramulator_trace()
        self.request_queue_size = self.config.get_req_buf_sz_rd()
        self.bw = self.config.get_bandwidths_as_list()[0]
        if self.ramulator_trace == True:
            self.latency_matrix = np.load(latency_file)
            #print(f"Latency file is {latency_file}")
        self.stall_cycles=0
        self.latency = 1
        
    def set_params(self, latency):
        """
        Method to set the backing buffer hit latency for housekeeping.
        """
        self.latency = latency

    #
    def get_latency(self):
        """
        Method to get the backing buffer hit latency for housekeeping.
        """
        return self.latency
    
    def find_latency(self):
        """
        Method to map DRAM return path latency for each transactions.
        """
        if(self.count < len(self.latency_matrix)):
            latency_out = self.latency_matrix[self.count]
            self.count+=1
        else:
            latency_out = self.latency
        if(latency_out > 10000):
            latency_out = 1
        return latency_out

    # The incoming read requests will be needed when the capability of port is expanded
    # At the moment its kept for compatibility
    def service_reads(self, incoming_requests_arr_np, incoming_cycles_arr):
        """
        Method to service read request by the read buffer.
        Check for hit in the request queue or add the DRAM
        roundtrip latency for each transaction reported by 
        Ramulator.
        """
        if self.ramulator_trace is False:
            out_cycles_arr = incoming_cycles_arr + self.latency
            return out_cycles_arr

        updated_req_timestamp = incoming_cycles_arr[0]
        out_cycles_arr = np.zeros(incoming_requests_arr_np.shape[0])
        for i in range(len(incoming_cycles_arr)):
            out_cycles_arr[i] = incoming_cycles_arr[i] + self.stall_cycles + self.find_latency()
            #print(str(incoming_cycles_arr[i]) + ' ' + str(out_cycles_arr[i]) + ' ' +str(self.stall_cycles))
            self.request_array.append(out_cycles_arr[i])
            if len(self.request_array) == self.request_queue_size:
                updated_req_timestamp = incoming_cycles_arr[i] + self.stall_cycles
                self.request_array.sort()
                if self.request_array[0] >= updated_req_timestamp:
                    self.stall_cycles += self.request_array[0] - updated_req_timestamp
                    #print(f"stall cycle: {self.stall_cycles}. request array: {self.request_array[0]} updated_req_timestamp: {updated_req_timestamp}")
                    updated_req_timestamp = self.request_array[0]
                    self.request_array.pop(0)
                else:
                    index = bisect_left(self.request_array,updated_req_timestamp)
                    if index == len(self.request_array):
                        self.request_array = []
                    else:
                        self.request_array = self.request_array[index:]
            elif len(self.request_array) > self.request_queue_size:
                self.request_array = self.request_array[-self.request_queue_size:]
                #print(f"stall cycle: {self.stall_cycles}. request array: {self.request_array[0]} updated_req_timestamp: {updated_req_timestamp}")
        
        self.stall_cycles=0
        return out_cycles_arr


"""
This module implements the 'systolic_compute_is' class, which simulates a systolic array with Weight
Stationary (WS) dataflow. It handles operand prefetching, demand matrix creation, and performance
metrics such as mapping efficiency and compute utilization. It also tracks read and write requests
for IFMAP, Filter, and OFMAP operations.
"""

import math
import numpy as np
from tqdm import tqdm
from scalesim.scale_config import scale_config as cfg
from scalesim.compute.compression import compression as cp

class systolic_compute_ws:
    """
    Class that computes the output using Weight Stationary dataflow.
    """
    #
    def __init__(self):
        """
        __init__ method.
        """
        # Params set by user
        self.config = cfg()
        self.sparsity_ratio_N = 1
        self.sparsity_ratio_M = 1

        self.ifmap_op_mat = np.zeros((1, 1))
        self.ofmap_op_mat = np.zeros((1, 1))
        self.filter_op_mat = np.zeros((1, 1))

        # Derived parameters
        self.Sr = 0
        self.Sc = 0
        self.T = 0

        self.arr_row = 0
        self.arr_col = 0

        self.row_fold = 1
        self.col_fold = 1

        # Generated matrices
        self.ifmap_op_mat_trans = np.zeros((1,1))
        self.ifmap_prefetch_matrix = np.zeros((1,1))
        self.filter_prefetch_matrix = np.zeros((1,1))

        self.ifmap_demand_matrix = np.zeros((1,1))
        self.ofmap_demand_matrix = np.zeros((1,1))
        self.filter_demand_matrix = np.zeros((1,1))

        self.ifmap_op_mat_original = np.zeros((1,1))
        self.sparsity_filter_array = np.zeros((1,1))

        # Generated metrics
        self.ifmap_reads = 0
        self.filter_reads = 0
        self.ofmap_writes = 0

        self.mapping_efficiency_per_fold = []
        self.compute_utility_per_fold = []

        # Flags
        self.params_set_flag = False
        self.prefetch_mat_ready_flag = False
        self.demand_mat_ready_flag = False

        # Compression
        self.compression = cp()

    #
    def set_params(self,
                   config_obj=cfg(),
                   ifmap_op_mat = np.zeros((1,1)),
                   ofmap_op_mat = np.zeros((1,1)),
                   filter_op_mat = np.zeros((1,1)),
                   sparsity_ratio_N = 1,
                   sparsity_ratio_M = 1,
                   ifmap_op_mat_original = np.zeros((1,1)),
                   sparsity_filter_array = np.zeros((1,1))
                ):
        """
        Method to set the weight stationary run parameters for housekeeping.
        """

        self.config = config_obj
        self.ifmap_op_mat = ifmap_op_mat
        self.filter_op_mat = filter_op_mat
        self.ofmap_op_mat = ofmap_op_mat
        self.sparsity_ratio_N = sparsity_ratio_N
        self.sparsity_ratio_M = sparsity_ratio_M

        self.ifmap_op_mat_original = ifmap_op_mat_original
        self.sparsity_filter_array = sparsity_filter_array

        ifmap_col = self.ifmap_op_mat.shape[1]
        filter_row = self.filter_op_mat.shape[0]

        if not self.config.sparsity_support and not self.config.sparsity_optimized_mapping:
            assert ifmap_col == filter_row, "Dimension mismatch between operands"

        self.Sr = self.ifmap_op_mat.shape[1]
        self.Sc = self.filter_op_mat.shape[1]
        self.T = self.ifmap_op_mat.shape[0]

        self.arr_row, self.arr_col = self.config.get_array_dims()

        self.row_fold = math.ceil(self.Sr / self.arr_row) # to be used for prefetch matrices
        self.row_fold_demand_matrices = math.ceil(self.filter_op_mat.shape[0] / self.arr_row)
        self.col_fold = math.ceil(self.Sc / self.arr_col)

        self.params_set_flag = True

    #
    def create_prefetch_matrices(self):
        """
        Method to create ifmap and filter prefetch matrices. These matrices are prefetched in the
        SRAM before running memory simulation.
        """
        assert self.params_set_flag, 'Parameters are not set'

        self.create_ifmap_prefetch_mat()
        self.create_filter_prefetch_mat()

        self.prefetch_mat_ready_flag = True

    #
    def create_ifmap_prefetch_mat(self):
        """
        Method to create ifmap prefetch matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        for fr in range(self.row_fold):
            start_col_idx = fr * self.arr_row
            end_col_idx = min(start_col_idx + self.arr_row, self.Sr)

            delta = self.arr_row - (end_col_idx - start_col_idx)

            this_fold_prefetch = self.ifmap_op_mat[:,start_col_idx: end_col_idx]

            #If there is under utilization, fill them with null requests
            if delta > 0:
                null_req_mat = np.ones((self.T, delta)) * -1
                this_fold_prefetch = np.concatenate((this_fold_prefetch, null_req_mat), axis=1)

            if fr == 0:
                self.ifmap_prefetch_matrix = this_fold_prefetch
            else:
                self.ifmap_prefetch_matrix = \
                    np.concatenate((self.ifmap_prefetch_matrix, this_fold_prefetch), axis=0)

        # Fixing ISSUE #15, #16
        # Roll out the matrices along the diagonal to account for temporal locality when there is a
        # skew in demand

        M, N = self.ifmap_prefetch_matrix.shape
        num_elems = M * N
        num_diags = M + N
        prefetches = np.zeros((1, num_elems))
        idx = 0

        pbar = tqdm(total=M * N, disable=True)
        # print('DEBUG: Total = ' + str(num_elems) + ' Diags = ' + str(num_diags))

        for diag_id in range(num_diags):
            max_row_id = min(diag_id, M - 1)
            min_row_id = max(0, diag_id - N + 1)
            valid_rows = max_row_id - min_row_id + 1

            for offset in range(valid_rows):
                row_id = max_row_id - offset
                col_id = diag_id - row_id

                elem = self.ifmap_prefetch_matrix[row_id][col_id]
                prefetches[0, idx] = elem
                idx += 1
                pbar.update(1)

        pbar.close()
        self.ifmap_prefetch_matrix = prefetches

    #
    def create_filter_prefetch_mat(self):
        """
        Method to create filter prefetch matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        for fc in range(self.col_fold):
            col_start_id = fc * self.arr_col
            col_end_id = min(col_start_id + self.arr_col, self.Sc)

            delta = self.arr_col - (col_end_id - col_start_id)

            this_fold_prefetch = self.filter_op_mat[:,col_start_id:col_end_id]

            if delta > 0:
                null_req_mat = np.ones((self.filter_op_mat.shape[0], delta)) * -1 # self.Sr
                this_fold_prefetch = np.concatenate((this_fold_prefetch, null_req_mat), axis=1)

            if fc == 0:
                self.filter_prefetch_matrix = this_fold_prefetch
            else:
                self.filter_prefetch_matrix = \
                    np.concatenate((self.filter_prefetch_matrix, this_fold_prefetch), axis=0)

        # Note: ISSUE #15: no skewing happens in the Filter for WS so this issue does not apply.

    #
    def create_demand_matrices(self):
        """
        Method to create ifmap, filter and ofmap demand matrices from the operand matrices. They
        contain several folds of ifmap, filter and ofmap demands. The folding happens because
        operand matrices are generally larger than systolic array dimensions.
        """
        # NCBS: check this once, create new assert for row_stationary, if...else
        assert self.params_set_flag, 'Parameters are not set'

        self.create_ifmap_demand_mat()
        self.create_filter_demand_mat()
        self.create_ofmap_demand_mat()   

        # assert self.ifmap_demand_matrix.shape[0] == self.filter_demand_matrix.shape[0], \
        #        'IFMAP and Filter demands out of sync'
        # assert self.ofmap_demand_matrix.shape[0] == self.filter_demand_matrix.shape[0], \
        #        'OFMAP and Filter demands out of sync'
        if not self.config.sparsity_optimized_mapping:
            assert self.ifmap_demand_matrix.shape[1] == self.arr_row, 'IFMAP demands exceed the rows'
        assert self.filter_demand_matrix.shape[1] == self.arr_col,'Filter demands exceed the cols'
        assert self.ofmap_demand_matrix.shape[1] == self.arr_col, 'OFMAP demands exceed the cols'

        self.demand_mat_ready_flag = True

    #
    def create_ifmap_demand_mat(self):
        """
        Method to create ifmap demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        inter_fold_gap_prefix = self.arr_row
        inter_fold_gap_prefix_mat = np.ones((inter_fold_gap_prefix, self.arr_row)) * -1

        inter_fold_gap_suffix = self.arr_col - 1

        inter_fold_gap_suffix_mat = np.ones((inter_fold_gap_suffix, self.arr_row)) * -1

        metadata_conversion_mat = [ [ ] ]
        if False:
            if self.config.sparsity_support is True:
                if self.config.sparsity_representation == 'csr':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'csc':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'ellpack_block':
                    metadata_conversion_mat = np.ones((0, self.arr_col)) * -1

        ifmap_demand_matrix_list = []

        for fc in range(self.col_fold):
            # for fr in range(self.row_fold):
            for fr in range(self.row_fold_demand_matrices):
                if self.config.sparsity_support and self.config.sparsity_optimized_mapping:
                    col_start_id = fr * (self.arr_row * 2) # Since we need 2 tiles
                    col_end_idx = min(col_start_id + (self.arr_row * 2), self.Sr)
                    delta = (self.arr_row * 2) - (col_end_idx - col_start_id)
                    this_fold_demand = self.ifmap_op_mat_original[:,col_start_id: col_end_idx]
                else:
                    col_start_id = fr * self.arr_row
                    col_end_idx = min(col_start_id + self.arr_row, self.Sr)
                    delta = self.arr_row - (col_end_idx - col_start_id)

                    # Indexing the cols with row start and row end idx are correct
                    # See the comment on ifmap_prefetch generation
                    this_fold_demand = self.ifmap_op_mat[:,col_start_id: col_end_idx]

                # Need to add custom skew for row-wise sparsity
                if self.config.sparsity_support and self.config.sparsity_optimized_mapping:
                    this_fold_demand = skew_matrix_row_sparsity(this_fold_demand, self.arr_row, \
                                                                self.config.sparsity_block_size)
                
                if self.config.sparsity_support:
                    if self.config.sparsity_optimized_mapping:
                        self.ifmap_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1]
                    else:
                        # A single block of input is shared among M/N rows, hence a row needs to be
                        # read M/N times (assume absence of any broadcast)
                        self.ifmap_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1] * \
                                            (self.sparsity_ratio_M / self.sparsity_ratio_N)
                else:
                    self.ifmap_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Take into account under utilization
                if not self.config.sparsity_optimized_mapping:
                    if delta > 0:
                        null_req_mat = np.ones((self.T, delta)) * -1
                        this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)
                
                if self.config.sparsity_support and self.config.sparsity_optimized_mapping:
                    if inter_fold_gap_prefix_mat.shape[1] < this_fold_demand.shape[1]:
                        inter_fold_gap_prefix_mat = np.pad(
                            inter_fold_gap_prefix_mat,
                            ((0, 0), (0, this_fold_demand.shape[1] - inter_fold_gap_prefix_mat.shape[1])),  # Pad only columns, not rows
                            constant_values=-1
                        )

                # Account for the cycles for weights to load
                this_fold_demand = np.concatenate((inter_fold_gap_prefix_mat, this_fold_demand),
                                                  axis=0)

                # Account for the cycles for final output to drain out
                if self.config.sparsity_support and self.config.sparsity_optimized_mapping:
                    if inter_fold_gap_suffix_mat.shape[1] < this_fold_demand.shape[1]:
                        inter_fold_gap_suffix_mat = np.pad(
                            inter_fold_gap_suffix_mat,
                            ((0, 0), (0, this_fold_demand.shape[1] - inter_fold_gap_suffix_mat.shape[1])),  # Pad only columns, not rows
                            constant_values=-1
                        )
                this_fold_demand = np.concatenate((this_fold_demand, inter_fold_gap_suffix_mat),
                                                  axis=0)                

                # Add skew to the IFMAP demand matrix to reflect systolic pipeline fill
                if not self.config.sparsity_optimized_mapping:
                    this_fold_demand = skew_matrix(this_fold_demand)

                ifmap_demand_matrix_list.append(this_fold_demand)

        self.ifmap_demand_matrix = np.concatenate(ifmap_demand_matrix_list)

        if False:
            if self.config.sparsity_support is True:
                self.ifmap_demand_matrix = \
                    np.concatenate((metadata_conversion_mat, self.ifmap_demand_matrix), axis=0)

    # END of IFMAP demand generation

    #
    def create_filter_demand_mat(self):
        """
        Method to create filter demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        inter_fold_gap_suffix = self.arr_row + self.arr_col + self.T - 2
        inter_fold_gap_suffix_mat = np.ones((inter_fold_gap_suffix, self.arr_col)) * -1

        metadata_conversion_mat = [ [ ] ]
        if False:
            if self.config.sparsity_support is True:
                if self.config.sparsity_representation == 'csr':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'csc':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'ellpack_block':
                    metadata_conversion_mat = np.ones((0, self.arr_col)) * -1

        filter_demand_matrix_list = []
        for fc in range(self.col_fold):
            # for fr in range(self.row_fold):
            for fr in range(self.row_fold_demand_matrices):
                row_start_id = fr * self.arr_row
                # row_end_idx = min(row_start_id + self.arr_row, self.Sr)
                row_end_idx = min(row_start_id + self.arr_row, self.filter_op_mat.shape[0])
                row_delta = self.arr_row - (row_end_idx - row_start_id)

                col_start_id = fc * self.arr_col
                col_end_idx = min(col_start_id + self.arr_col, self.Sc)
                col_delta = self.arr_col - (col_end_idx - col_start_id)

                this_fold_demand = \
                    self.filter_op_mat[row_start_id:row_end_idx, col_start_id: col_end_idx]
                self.filter_reads += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Take into account under utilization
                if col_delta > 0:
                    null_req_mat = np.ones((this_fold_demand.shape[0], col_delta)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)

                if row_delta > 0:
                    null_req_mat = np.ones((row_delta, self.arr_col)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=0)

                # The filters are needed to be filled in reverse order to ensure that
                # top element is pushed in last to maintain alignment with the input elements
                this_fold_demand = np.flip(this_fold_demand, 0)

                sum_sparse = sum(list(row).count(-1) for row in this_fold_demand)

                # Time for inputs to stream and the partial sums to drain out
                this_fold_demand = np.concatenate((this_fold_demand, inter_fold_gap_suffix_mat),
                                                  axis=0)

                # Calculate the mapping efficiency
                row_used = min(self.arr_row, row_end_idx - row_start_id)
                col_used = min(self.arr_col, col_end_idx - col_start_id)
                mac_used = row_used * col_used

                # mapping_eff_this_fold = mac_used / (self.arr_row * self.arr_col)
                mapping_eff_this_fold = \
                    ((self.arr_row * self.arr_col) - sum_sparse) / (self.arr_row * self.arr_col)

                cycles_this_fold = this_fold_demand.shape[0] + this_fold_demand.shape[1] - 1
                compute_cycles_this_fold = mac_used * self.T
                compute_util_this_fold = \
                    compute_cycles_this_fold / (self.arr_row * self.arr_col * cycles_this_fold)

                self.mapping_efficiency_per_fold.append(mapping_eff_this_fold)
                self.compute_utility_per_fold.append(compute_util_this_fold)

                filter_demand_matrix_list.append(this_fold_demand)
                #if fr == 0 and fc == 0:
                #    self.filter_demand_matrix = this_fold_demand
                #else:
                #    self.filter_demand_matrix = \
                #       np.concatenate((self.filter_demand_matrix, this_fold_demand), axis=0)
        
        self.filter_demand_matrix = np.concatenate(filter_demand_matrix_list)

        if False:
            if self.config.sparsity_support is True:
                self.filter_demand_matrix = \
                    np.concatenate((metadata_conversion_mat, self.filter_demand_matrix), axis=0)

        # No skew needed in filters for weight stationary

    #
    def create_ofmap_demand_mat(self):
        """
        Method to create ofmap demand matrix.
        """
        assert self.params_set_flag, 'Parameters are not set'

        inter_fold_gap_prefix = 2 * self.arr_row - 1
        inter_fold_gap_prefix_mat = np.ones((inter_fold_gap_prefix, self.arr_col)) * -1

        metadata_conversion_mat = [ [ ] ]
        if False:
            if self.config.sparsity_support is True:
                if self.config.sparsity_representation == 'csr':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'csc':
                    metadata_conversion_mat = np.ones((1, self.arr_col)) * -1
                elif self.config.sparsity_representation == 'ellpack_block':
                    metadata_conversion_mat = np.ones((0, self.arr_col)) * -1

        ofmap_demand_matrix_list = []

        for fc in range(self.col_fold):
            # for fr in range(self.row_fold):
            for fr in range(self.row_fold_demand_matrices):
                col_start_id = fc * self.arr_col
                col_end_idx = min(col_start_id + self.arr_col, self.Sc) # self.Sc
                col_delta = self.arr_col - (col_end_idx - col_start_id)

                this_fold_demand = self.ofmap_op_mat[:, col_start_id: col_end_idx]
                self.ofmap_writes += this_fold_demand.shape[0] * this_fold_demand.shape[1]

                # Adding null requests when there is under utilization ie. no mapping along a few
                # rows or cols
                if col_delta > 0:
                    null_req_mat = np.ones((this_fold_demand.shape[0], col_delta)) * -1
                    this_fold_demand = np.concatenate((this_fold_demand, null_req_mat), axis=1)

                # Now add the prefix matrix
                # These are the null demands to account for when the operands are streamed in
                # and the OFMAPS are not ready
                this_fold_demand = np.concatenate((inter_fold_gap_prefix_mat, this_fold_demand),
                                                  axis=0)

                # Add skew to the OFMAP demand matrix to reflect systolic pipeline fill
                this_fold_demand = skew_matrix(this_fold_demand)

                ofmap_demand_matrix_list.append(this_fold_demand)
                #if fr == 0 and fc == 0:
                #    self.ofmap_demand_matrix = this_fold_demand
                #else:
                #    self.ofmap_demand_matrix = \
                #       np.concatenate((self.ofmap_demand_matrix, this_fold_demand), axis=0)

        self.ofmap_demand_matrix = np.concatenate(ofmap_demand_matrix_list)

        if False:
            if self.config.sparsity_support is True:
                self.ofmap_demand_matrix = \
                    np.concatenate((metadata_conversion_mat, self.ofmap_demand_matrix), axis=0)

    # END of OFMAP demand generation

    #
    def get_ifmap_prefetch_mat(self):
        """
        Method to get ifmap prefetch matrix.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.ifmap_prefetch_matrix

    #
    def get_filter_prefetch_mat(self):
        """
        Method to get filter prefetch matrix.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.filter_prefetch_matrix

    #
    def get_prefetch_matrices(self):
        """
        Method to get ifmap and filter prefetch matrices.
        """
        if not self.prefetch_mat_ready_flag:
            self.create_prefetch_matrices()

        return self.ifmap_prefetch_matrix, self.filter_prefetch_matrix

    #
    def get_ifmap_demand_mat(self):
        """
        Method to get ifmap demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ifmap_demand_matrix

    #
    def get_filter_demand_mat(self):
        """
        Method to get filter demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.filter_demand_matrix

    #
    def get_ofmap_demand_mat(self):
        """
        Method to get ofmap demand matrix.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ofmap_demand_matrix

    #
    def get_demand_matrices(self):
        """
        Method to get ifmap, filter and ofmap demand matrices.
        """
        if not self.demand_mat_ready_flag:
            self.create_demand_matrices()

        return self.ifmap_demand_matrix, self.filter_demand_matrix, self.ofmap_demand_matrix

    #
    def get_avg_mapping_efficiency(self):
        """
        Method to get average mapping efficincy on the systolic array.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'

        agg = sum(self.mapping_efficiency_per_fold)
        num = len(self.mapping_efficiency_per_fold)

        avg_mapping_eff = agg / num

        return avg_mapping_eff

    #
    def get_avg_compute_utilization(self):
        """
        Method to get average compute utilization on the systolic array.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'

        agg = sum(self.compute_utility_per_fold)
        num = len(self.compute_utility_per_fold)

        avg_compute_util = agg / num

        return avg_compute_util

    #
    def get_ifmap_requests(self):
        """
        Method to get the number of ifmap read requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.ifmap_reads

    #
    def get_filter_requests(self):
        """
        Method to get the number of filter read requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.filter_reads

    #
    def get_ofmap_requests(self):
        """
        Method to get the number of ofmap read requests.
        """
        assert self.demand_mat_ready_flag, 'Computes not ready yet'
        return self.ofmap_writes


#
def skew_matrix(input_matrix_np):
    """
    Method to add skew to the input matix to maintain systolic array flow.
    Example:
        Input matrix:
        1 1 1 1 1 1 1 1 1

        Output matrix:
            1 1 1
          1 1 1
        1 1 1
    """
    rows, cols = input_matrix_np.shape

    out_matrix_np = np.full((rows + cols - 1, cols), -1, dtype=input_matrix_np.dtype)

    for c in range(cols):
        out_matrix_np[c:c + rows, c] = input_matrix_np[:, c]

    return out_matrix_np

#
def skew_matrix_row_sparsity(input_matrix, arr_row, block_size):
    # Step 1: Ensure the number of columns is arr_row * 2 as we are combining 2 tiles
    num_tiles = 2
    num_cols = input_matrix.shape[1]
    if num_cols < arr_row * num_tiles:
        padding = arr_row * num_tiles - num_cols
        input_matrix = np.pad(input_matrix, ((0, 0), (0, padding)), constant_values=-1)
    
    # Step 2: Split into blocks
    blocks = []
    for row in input_matrix:
        row_blocks = [row[i:i+block_size] for i in range(0, len(row), block_size)]
        blocks.append(row_blocks)

    # Step 3: Add extra copies to match block_size / 2 copies per column
    repeated_blocks = []
    for block_row in blocks:
        new_row = []
        for block in block_row:
            new_row.extend([block] * (block_size // num_tiles))
        repeated_blocks.append(new_row)

    # Step 4: Apply skew to the repeated blocks
    output_matrix = []
    num_block_rows = len(repeated_blocks)
    num_block_cols = (len(repeated_blocks[0][0]) * block_size) // num_tiles

    for i in range(num_block_rows + arr_row - 1):
        row = []
        for j in range(len(repeated_blocks[0])):
            block_row_idx = i - j
            if 0 <= block_row_idx < num_block_rows:
                row.append(repeated_blocks[block_row_idx][j])
            else:
                row.append([-1] * block_size)
        row = np.concatenate(row)
        output_matrix.append(row)
    
    return np.array(output_matrix, dtype=input_matrix.dtype)



"""
This file contains the 'topologies' class that handles the topology files fed to SCALE_Sim tool.
"""

import math


class topologies(object):
    """
    Class which contains the methods to preprocess the data from topology file (.csv format) before
    doing compute simulation.
    """
    #
    def __init__(self):
        """
        __init__ method
        """
        self.current_topo_name = ""
        self.topo_file_name = ""
        self.topo_arrays = []
        self.spatio_temp_dim_arrays = []
        self.layers_calculated_hyperparams = []
        self.num_layers = 0
        self.topo_load_flag = False
        self.topo_calc_hyper_param_flag = False
        self.topo_calc_spatiotemp_params_flag = False
        self.df = ""
        self.current_toponame = ""
        self.layer_name = ""

    # reset topology parameters
    def reset(self):
        """
        Method to reset the topology parameters.
        """
        print("All data reset")
        self.current_topo_name = ""
        self.topo_file_name = ""
        self.topo_load_flag = False
        self.topo_arrays = []
        self.num_layers = 0
        self.topo_calc_hyper_param_flag = False
        self.layers_calculated_hyperparams = []
        self.df = ""
        self.current_toponame = ""
        self.layer_name = ""

    #
    def load_layer_params_from_list(self, layer_name, elems_list=[]):
        """
        Method to load layer parameters from the given layer name and element list.
        """
        self.topo_file_name = ''
        self.current_toponame = ''
        self.layer_name = layer_name
        self.append_topo_arrays(layer_name, elems_list)

        self.num_layers += 1
        self.topo_load_flag = True

    #
    def load_arrays(self, topofile='', mnk_inputs=False):
        """
        Method to read the topology file and collect names and dimensions of all the workload
        layers.
        """
        if mnk_inputs:
            self.load_arrays_gemm(topofile)
        else:
            self.load_arrays_conv(topofile)

    #
    def load_arrays_gemm(self, topofile=''):
        """
        Method to read the GEMM topology file and collect names and dimensions of all the workload
        layers.
        """

        self.topo_file_name = topofile.split('/')[-1]
        name_arr = self.topo_file_name.split('.')
        if len(name_arr) > 1:
            self.current_topo_name = self.topo_file_name.split('.')[-2]
        else:
            self.current_topo_name = self.topo_file_name

        f = open(topofile, 'r')
        first = True

        for row in f:
            row = row.strip()
            if first:
                first = False
                continue
            elif row == '':
                continue
            else:
                elems = row.split(',')[:-1]
                assert len(elems) > 3, 'There should be at least 4 entries per row'
                layer_name = elems[0].strip()
                m = elems[1].strip()
                n = elems[2].strip()
                k = elems[3].strip()
                if len(elems) < 5:
                    # If sparsity ratio is missing in the topology file, consider the default ratio
                    elems.append("1:1")
                sparsity_ratio = elems[4].strip().split(':')

                # Entries: layer name, Ifmap h, ifmap w, filter h, filter w, num_ch, num_filt,
                #          stride h, stride w, N in N:M, M in N:M
                entries = [layer_name, m, k, 1, k, 1, n, 1, 1, sparsity_ratio[0], sparsity_ratio[1]]
                # entries are later iterated from index 1. Index 0 is used to store layer name in
                # convolution mode. So, to rectify assignment of M, N and K in GEMM mode, layer name
                # has been added at index 0 of entries.
                self.append_topo_arrays(layer_name=layer_name, elems=entries)

        self.num_layers = len(self.topo_arrays)
        self.topo_load_flag = True

    # Load the topology data from the file
    def load_arrays_conv(self, topofile=''):
        """
        Method to read the CONV topology file and collect names and dimensions of all the workload
        layers.
        """
        first = True
        self.topo_file_name = topofile.split('/')[-1]
        name_arr = self.topo_file_name.split('.')
        if len(name_arr) > 1:
            self.current_topo_name = self.topo_file_name.split('.')[-2]
        else:
            self.current_topo_name = self.topo_file_name

        f = open(topofile, 'r')
        for row in f:
            row = row.strip()
            if first or row == '':
                first = False
            else:
                elems = row.split(',')[:-1]

                # Add the same stride in the col direction automatically
                elems = elems[0:8] + [elems[7]] + elems[8:]

                # Parsing sparsity ratio
                if len(elems) < 10:
                    # If sparsity ratio is missing in the topology file, consider the default ratio
                    elems.append("1:1")
                sparsity_ratio = elems.pop().strip().split(':')
                elems.append(sparsity_ratio[0])
                elems.append(sparsity_ratio[1])

                # depth-wise convolution
                if 'DP' in elems[0].strip():
                    for dp_layer in range(int(elems[5].strip())):
                        layer_name = elems[0].strip() + "Channel_" + str(dp_layer)
                        elems[5] = str(1)
                        self.append_topo_arrays(layer_name, elems)
                else:
                    layer_name = elems[0].strip()
                    self.append_topo_arrays(layer_name, elems)

        self.num_layers = len(self.topo_arrays)
        self.topo_load_flag = True

    # Write the contents into a csv file
    def write_topo_file(self,
                      path="",
                      filename=""
                      ):
        """
        Method to write the workload data into a csv file.
        """
        if path == "":
            print("WARNING: topology_utils.write_topo_file: No path specified writing to the cwd")
            path = "./"

        if filename == "":
            print("ERROR: topology_utils.write_topo_file: No filename provided")
            return

        filename = path + "/" + filename

        if not self.topo_load_flag:
            print("ERROR: topology_utils.write_topo_file: No data loaded")
            return

        header = [
                    "Layer name",
                    "IFMAP height",
                    "IFMAP width",
                    "Filter height",
                    "Filter width",
                    "Channels",
                    "Num filter",
                    "Stride height",
                    "Stride width"
                ]

        f = open(filename, 'w')
        log = ",".join(header)
        log += ",\n"
        f.write(log)

        for param_arr in self.topo_arrays:
            log = ",".join([str(x) for x in param_arr])
            log += ",\n"
            f.write(log)

        f.close()

    # LEGACY
    def append_topo_arrays(self, layer_name, elems):
        """
        Method to append the layer dimensions in int data type and layer name to the topo_arrays
        variable. This method also checks that the filter dimensions do not exceed the ifmap
        dimensions.
        """
        entry = [layer_name]

        for i in range(1, len(elems)):
            val = int(str(elems[i]).strip())
            entry.append(val)
            # These 3 lines are taken care in the parent function call
            # if i == 7 and len(elems) < 9:
            #     print("Adding extra value")
            #     entry.append(val)  # Add the same stride in the col direction automatically

        # ISSUE #9 Fix
        assert entry[3] <= entry[1], 'Filter height cannot be larger than IFMAP height'
        assert entry[4] <= entry[2], 'Filter width cannot be larger than IFMAP width'

        self.topo_arrays.append(entry)

    # create network topology array
    def append_topo_entry_from_list(self, layer_entry_list=[]):
        """
        Method to append the layer dimensions in int data type and layer name.
        """
        assert 7 < len(layer_entry_list) < 10, 'Incorrect number of parameters'

        entry = [str(layer_entry_list[0])]

        for i in range(1, len(layer_entry_list)):
            val = int(str(layer_entry_list[i]).strip())
            entry.append(val)
            if i == 7 and len(layer_entry_list) < 9:
                entry.append(val)  # Add the same stride in the col direction automatically

        self.append_layer_entry(entry,toponame=self.current_topo_name)

    # add to the existing data from a list
    def append_layer_entry(self, entry, toponame=""):
        """
        Method to append data of a single layer to the array containing data of all the layers. This
        method also calls topo_calc_hyperparams method to calculate the hyperparameters.
        """
        assert len(entry) == 9, 'Incorrect number of parameters'

        if toponame != "":
            self.current_topo_name = toponame

        self.topo_arrays.append(entry)
        self.topo_load_flag = True
        self.topo_calc_hyperparams()
        self.num_layers += 1

    # calculate hyper-parameters (ofmap dimensions, number of MACs, and window size of filter)
    def topo_calc_hyperparams(self, topofilename=""):
        """
        Method to calculate hyper-parameters (ofmap dimensions, number of MACs, and window size of
        filter) if topology array is loaded.
        """
        if not self.topo_load_flag:
            self.load_arrays(topofilename)
        self.layers_calculated_hyperparams = []
        for array in self.topo_arrays:
            ifmap_h = array[1]
            ifmap_w = array[2]
            filt_h = array[3]
            filt_w = array[4]
            num_ch   = array[5]
            num_filt = array[6]
            stride_h = array[7]
            stride_w = array[8]
            ofmap_h = int(math.ceil((ifmap_h - filt_h + stride_h) / stride_h))
            ofmap_w = int(math.ceil((ifmap_w - filt_w + stride_w) / stride_w))
            num_mac = ofmap_h * ofmap_w * filt_h * filt_w * num_ch * num_filt
            window_size = filt_h * filt_w * num_ch
            entry = [ofmap_h, ofmap_w, num_mac, window_size]
            self.layers_calculated_hyperparams.append(entry)
        self.topo_calc_hyper_param_flag = True

    #
    def calc_spatio_temporal_params(self, df='os', layer_id=0):
        """
        Method to calculate spatio-temporal parameters (S_r, S_c and T) based on the dataflow.
        (Refer the scalesim paper for more info)
        """
        s_row = -1
        s_col = -1
        t_time = -1
        if self.topo_calc_hyper_param_flag:
            num_filt  = self.get_layer_num_filters(layer_id= layer_id)
            num_ofmap = self.get_layer_num_ofmap_px(layer_id=layer_id)
            num_ofmap = int(num_ofmap / num_filt)
            window_sz = self.get_layer_window_size(layer_id=layer_id)
            if df == 'os':
                s_row = num_ofmap
                s_col = num_filt
                t_time = window_sz
            elif df == 'ws':
                s_row = window_sz
                s_col = num_filt
                t_time = num_ofmap
            elif df == 'is':
                s_row = window_sz
                s_col = num_ofmap
                t_time = num_filt
        else:
            self.topo_calc_hyperparams(self.topo_file_name)
        return s_row, s_col, t_time

    #
    def set_spatio_temporal_params(self):
        """
        Method to calculate spatio-temporal parameters (S_r, S_c and T) for all the layers
        """
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams(self.topo_file_name)
        for i  in range(self.num_layers):
            this_layer_params_arr = []
            for df in ['os', 'ws', 'is']:
                sr, sc, tt = self.calc_spatio_temporal_params(df=df, layer_id=i)
                this_layer_params_arr.append([sr, sc, tt])
            self.spatio_temp_dim_arrays.append(this_layer_params_arr)
        self.topo_calc_spatiotemp_params_flag = True

    #
    def get_transformed_mnk_dimensions(self):
        """
        Method to get M, N and K parameters for all the layers. These are GEMM parameters in which
        an input matrix of dimensions MxN is multiplied to a filter matrix of dimensions NxK.
        """
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams(self.topo_file_name)

        mnk_dims_arr = []
        for i in range(self.num_layers):
            M = self.get_layer_num_ofmap_px(layer_id=i)
            N = self.get_layer_num_filters(layer_id=i)
            K = self.get_layer_window_size(layer_id=i)

            mnk_dims_arr.append([M, N, K])

        return mnk_dims_arr

    #
    def get_current_topo_name(self):
        """
        Method to get the name of the workload if available. If not, print an error message.
        """
        current_topo_name = ""
        if self.topo_load_flag:
            current_topo_name = self.current_topo_name
        else:
            print('Error: get_current_topo_name(): Topo file not read')
        return current_topo_name

    #
    def get_num_layers(self):
        """
        Method to get the number of layers of the workload if available. If not, print an error
        message.
        """
        if not self.topo_load_flag:
            print("ERROR: topologies.get_num_layers: No array loaded")
            return
        return self.num_layers

    #
    def get_layer_ifmap_dims(self, layer_id=0):
        """
        Method to get the ifmap dimensions of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_ifmap_dims: Invalid layer id")

        layer_params = self.topo_arrays[layer_id]
        return layer_params[1:3]    # Idx = 1, 2

    #
    def get_layer_filter_dims(self, layer_id=0):
        """
        Method to get the filter dimensions of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_ifmap_dims: Invalid layer id")

        layer_params = self.topo_arrays[layer_id]
        return layer_params[3:5]    # Idx = 3, 4

    #
    def get_layer_num_filters(self, layer_id=0):
        """
        Method to get the number of filters of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_num_filter: Invalid layer id")
        layer_params = self.topo_arrays[layer_id]
        return layer_params[6]

    #
    def get_layer_num_channels(self, layer_id=0):
        """
        Method to get the number of channels of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_num_filter: Invalid layer id")
        layer_params = self.topo_arrays[layer_id]
        return layer_params[5]

    #
    def get_layer_strides(self, layer_id=0):
        """
        Method to get the strides of the layer if available. If not, print an error message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_strides: Invalid layer id")

        layer_params = self.topo_arrays[layer_id]
        return layer_params[7:9]

    #
    def get_layer_sparsity_ratio(self, layer_id=0):
        """
        Method to get the sparsity ratio of the layer if available. If not, print an error message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_sparsity_ratio: Invalid layer id")

        layer_params = self.topo_arrays[layer_id]
        return layer_params[9:11]

    #
    def get_layer_window_size(self, layer_id=0):
        """
        Method to get the convolution window size of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_num_filter: Invalid layer id")
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams()
        layer_calc_params = self.layers_calculated_hyperparams[layer_id]
        return layer_calc_params[3]

    #
    def get_layer_num_ofmap_px(self, layer_id=0):
        """
        Method to get the number of ofmap pixels of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_num_filter: Invalid layer id")
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams()
        layer_calc_params = self.layers_calculated_hyperparams[layer_id]
        num_filters = self.get_layer_num_filters(layer_id)
        num_ofmap_px = layer_calc_params[0] * layer_calc_params[1] * num_filters
        return num_ofmap_px

    #
    def get_layer_ofmap_dims(self, layer_id=0):
        """
        Method to get the ofmap dimensions of the layer if available. If not, print an error
        message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_ofmap_dims: Invalid layer id")
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams()
        ofmap_dims = self.layers_calculated_hyperparams[layer_id][0:2]
        return ofmap_dims

    #
    def get_layer_params(self, layer_id=0):
        """
        Method to get the parameters of the layer if available. If not, print an error message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_params: Invalid layer id")
            return
        layer_params = self.topo_arrays[layer_id]
        return layer_params

    #
    def get_layer_id_from_name(self, layer_name=""):
        """
        Method to get layer number from the given layer name if available. If not, print an error
        message.
        """
        if (not self.topo_load_flag) or layer_name == "":
            print("ERROR")
            return
        indx = -1
        for i in range(len(self.topo_arrays)):
            if layer_name == self.topo_arrays[i]:
                indx = i
        if indx == -1:
            print("WARNING: Not found")
        return indx

    #
    def get_layer_name(self, layer_id=0):
        """
        Method to get the layer name from the given layer number if available. If not, print an
        error message.
        """
        if not (self.topo_load_flag or self.num_layers - 1 < layer_id):
            print("ERROR: topologies.get_layer_name: Invalid layer id")
            return

        name = self.topo_arrays[layer_id][0]
        return str(name)

    #
    def get_layer_names(self):
        """
        Method to get the names of all the layers in the workload. If not, print an error message.
        """
        if not self.topo_load_flag:
            print("ERROR")
            return
        layer_names = []
        for entry in self.topo_arrays:
            layer_name = str(entry[0])
            layer_names.append(layer_name)
        return layer_names

    #
    def get_layer_mac_ops(self, layer_id=0):
        """
        Method to get the number of mac operations of the layer if hyper-parameters are calculated.
        If not, calculate the hyper-parameters first.
        """
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams(topofilename=self.topo_file_name)
        layer_hyper_param = self.layers_calculated_hyperparams[layer_id]
        mac_ops = layer_hyper_param[2]
        return mac_ops

    #
    def get_all_mac_ops(self):
        """
        Method to get the total number of mac operations of all the layer if hyper-parameters are
        calculated. If not, calculate the hyper-parameters first.
        """
        if not self.topo_calc_hyper_param_flag:
            self.topo_calc_hyperparams(topofilename=self.topo_file_name)
        total_mac = 0
        for layer in range(self.num_layers):
            total_mac += self.get_layer_mac_ops(layer)
        return total_mac

    # spatio-temporal dimensions specific to dataflow
    def get_spatiotemporal_dims(self, layer_id=0, df=''):
        """
        Method to get the spatio-temporal dimensions (S_r, S_c, T) of the layer if spatio-temporal
        parameters are calculated. If not, calculate the spatio-temporal parameters first. (refer to
        the scalesim paper for more info)
        """
        if df == '':
            df = self.df
        if not self.topo_calc_spatiotemp_params_flag:
            self.set_spatio_temporal_params()
        df_list = ['os', 'ws', 'is']
        df_idx = df_list.index(df)
        s_row = self.spatio_temp_dim_arrays[layer_id][df_idx][0]
        s_col = self.spatio_temp_dim_arrays[layer_id][df_idx][1]
        t_time = self.spatio_temp_dim_arrays[layer_id][df_idx][2]
        return s_row, s_col, t_time


if __name__ == '__main__':
    tp = topologies()
